# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Storage definition credentials and related helpers.
"""

import base64
import os
from typing import Union

from .constants import *
from .exception import *
from .credentials import (
    Credential,
    CredentialAesKey,
    CredentialByteArray,
    CredentialStore,
    DescribedCredential,
    prompt_for_password_unlock_credential,
)


def get_credential_class(
    store_credential_name: str,
):
    # Determine the type needed by using the cred_def info.
    #   Backup encryption key: CredentialAesKey
    #   Storage secret: Credential

    cls = None
    if store_credential_name == CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION:
        cls = CredentialAesKey
    elif store_credential_name == CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD:
        cls = Credential
    else:
        raise ValueError(f"Unexpected: store_credential_name={store_credential_name}")
    return cls


@dataclass
class StorageDefCredential:
    desc_cred: DescribedCredential
    affected_section: dict
    affected_section_key: str


class StorageDefCredentialSet:
    def __init__(
        self,
        storage_def_name: str = None,
        storage_def_dict: dict = None,
    ) -> None:
        self.storage_def_name = storage_def_name
        self.storage_def_dict = storage_def_dict
        self.storage_def_creds = list[StorageDefCredential]()
    def clear(self):
        self.storage_def_creds.clear()
    def get_affected_section(
        self,
        affected_config_path_parts: Union[str, list[str]],
    ) -> tuple[dict, str]:
        if isinstance(affected_config_path_parts, str):
            # Split to coerce str into storage definition path parts. i.e.,
            #   'driver-secret' -> ['driver', 'secret']
            #   'encryption-key' -> ['encryption', 'key']
            affected_config_path_parts = affected_config_path_parts.split("-")
        if not isinstance(affected_config_path_parts, list):
            raise ValueError(
                f"affected_config_path_parts must be either a str or list of "
                f"section parts."
            )
        # Walk from top of storage_def down to the second-to-last section
        # which is where the secret goes.
        affected_section = self.storage_def_dict
        for section_part in affected_config_path_parts[:-1]:
            if section_part not in affected_section:
                affected_section[section_part] = {}
            affected_section = affected_section[section_part]
        return affected_section, affected_config_path_parts[-1]

    def populate(self):
        """For each credential associated with this set's storage definition,
        add the credential to the set.
        """
        if len(self.storage_def_creds) != 0:
            raise InvalidStateError(
                f"StorageDefCredentialSet.populate can only be called on an empty set."
            )
        for _, cred_def in CREDENTIAL_DEFINITIONS.items():

            (affected_section, affected_section_key,) = self.get_affected_section(
                affected_config_path_parts=cred_def.section_path,
            )

            if affected_section_key not in affected_section:
                continue

            raw_cred_value = affected_section[affected_section_key]

            if raw_cred_value == CONFIG_KEY_VALUE_KEYRING_INDIRECTION:
                desc_cred = CredentialStore().get_credential(
                    config_name=self.storage_def_name,
                    credential_name=cred_def.store_credential_name,
                )
            else:
                # Currently, the Credential* type is not persisted in the raw.
                # Determine class to use from store credential name.
                cls = get_credential_class(
                    store_credential_name=cred_def.store_credential_name,
                )
                credential = cls.create_credential_from_bytes(
                    cred_bytes=CredentialByteArray(
                        base64.b64decode(raw_cred_value.encode("utf-8"))
                    )
                )
                password_type = affected_section[CONFIG_PASSWORD_TYPE]
                desc_cred = DescribedCredential(
                    credential=credential,
                    config_name=self.storage_def_name,
                    credential_name=cred_def.store_credential_name,
                    credential_kind=password_type,
                )
            self.append(
                desc_cred=desc_cred,
                affected_config_path_parts=cred_def.section_path,
            )

    def append(
        self,
        desc_cred: DescribedCredential,
        affected_config_path_parts: Union[str, list[str]],
    ):
        (affected_section, affected_section_key,) = self.get_affected_section(
            affected_config_path_parts=affected_config_path_parts,
        )
        self.storage_def_creds.append(
            StorageDefCredential(
                desc_cred=desc_cred,
                affected_section=affected_section,
                affected_section_key=affected_section_key,
            )
        )

    def set_password(
        self,
        cred_password: CredentialByteArray,
    ):
        for storage_def_cred in self.storage_def_creds:
            credential = storage_def_cred.desc_cred.credential
            if credential.is_password_protected:
                if credential.is_password_required:
                    raise CredentialStateInvalid(
                        f"Cannot set password while credential is protected."
                    )
            if cred_password is not None:
                credential.prepare_for_new_password()
                credential.set(password=CredentialByteArray(cred_password))
            else:
                credential.clear_password_protection()

    def unprotect(
        self,
        base64_encoded_secrets: bool = False,
        password: CredentialByteArray = None,
    ):
        password_protected_status = None
        for storage_def_cred in self.storage_def_creds:
            credential = storage_def_cred.desc_cred.credential
            if not credential.is_password_protected:
                if password_protected_status is not None and password_protected_status:
                    # A previous credential in the set was password-protected.
                    # This is currently not a use case.
                    raise CredentialSetInvalid(
                        f"Credentials in the set have differing password protection status."
                    )
                password_protected_status = False
            else:
                if (
                    password_protected_status is not None
                    and not password_protected_status
                ):
                    # A previous credential in the set was *not* password-protected.
                    # This is currently not a use case.
                    raise CredentialSetInvalid(
                        f"Credentials in the set have differing password protection status."
                    )
                password_protected_status = True
                if password is None:
                    prompt_for_password_unlock_credential(
                        credential=credential,
                        prompt=f"Enter the password for '{self.storage_def_name}':"
                    )
                    password = credential.password
                else:
                    credential.set(password=CredentialByteArray(password))
                    credential.decrypt_key()
            secret_to_use = credential.the_key
            if base64_encoded_secrets:
                secret_to_use = base64.b64encode(
                    credential.get_material_as_bytes()
                ).decode("utf-8")
            storage_def_cred.affected_section[
                storage_def_cred.affected_section_key
            ] = secret_to_use
            storage_def_cred.affected_section[
                CONFIG_PASSWORD_TYPE
            ] = storage_def_cred.desc_cred.credential_kind

    def protect(self):
        password_protected_status = None
        for storage_def_cred in self.storage_def_creds:
            credential = storage_def_cred.desc_cred.credential
            cred_name = storage_def_cred.desc_cred.credential_name
            if not credential.is_password_protected:
                if password_protected_status is not None and password_protected_status:
                    # A previous credential in the set was password-protected.
                    # This is currently not a use case.
                    raise CredentialSetInvalid(
                        f"Credentials in the set have differing password protection status."
                    )
                password_protected_status = False
                continue
            if password_protected_status is not None and not password_protected_status:
                # A previous credential in the set was *not* password-protected.
                # This is currently not a use case.
                raise CredentialSetInvalid(
                    f"Credentials in the set have differing password protection status."
                )
            password_protected_status = True
            print(f"Encrypting key '{cred_name}'...", end="")
            storage_def_cred.desc_cred.credential.encrypt_key()
            print(f"encrypted.")
            credential.clear_password()

    def save(self):
        rollback_info = []
        try:
            for storage_def_cred in self.storage_def_creds:
                desc_cred = storage_def_cred.desc_cred
                cba_old = CredentialStore().set_credential(desc_cred=desc_cred)
                if cba_old is not None:
                    rollback_info.append(
                        (
                            desc_cred.config_name,
                            desc_cred.credential_name,
                            cba_old,
                        )
                    )
                storage_def_cred.affected_section[
                    storage_def_cred.affected_section_key
                ] = CONFIG_KEY_VALUE_KEYRING_INDIRECTION
        except BaseException as ex:
            print(f"Failure saving credentials: {ex}")
            try:
                if len(rollback_info) > 0:
                    print(f"Rolling back already writting changes...")
                for rbi in rollback_info:
                    print(f"Undoing: config_name={rbi[0]} credential_name={rbi[1]}")
                    CredentialStore().provider.set_cred_bytes(
                        config_name=rbi[0],
                        credential_name=rbi[1],
                        cred_ascii_bytes=rbi[2],
                    )
            except BaseException as ex2:
                print(f"Rollback error: {ex2}")
            raise

    def get_desc_cred(
        self,
        credential_name: str,
    ):
        for storage_def_cred in self.storage_def_creds:
            cred_name = storage_def_cred.desc_cred.credential_name
            if cred_name == credential_name:
                return storage_def_cred.desc_cred
        return None

    def get_encryption_desc_cred(self):
        return self.get_desc_cred(
            credential_name=CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION
        )

    def get_storage_desc_cred(self):
        return self.get_desc_cred(
            credential_name=CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD
        )


def prompt_user_password_file_does_not_exist(default_filename: str):
    is_exist = default_filename is not None and os.path.isfile(default_filename)
    print(f"Specify a path to an existing password .json file.")
    def_qual = ""
    if is_exist:
        def_qual = "<ENTER> for the default, "
        print(
            f"If you do not specify a name and press <ENTER>, the default name will be used."
        )
        print(f"Default path (file exists): {default_filename}")
    a = input(f"Enter the path, {def_qual}or 'n' to abort:")
    if a.lower() in ["n", "no"]:
        return None
    if is_exist and a == "":
        return default_filename
    return a


def restore_keyring_secrets(
    storage_def_name: str,
    storage_def: dict,
):
    for cred_def_friendly_name, cred_def in CREDENTIAL_DEFINITIONS.items():
        affected_config_path_parts = cred_def.section_path.split("-")
        affected_section = storage_def
        for section_name in affected_config_path_parts[:-1]:
            if section_name not in affected_section:
                affected_section = None
                break
            affected_section = affected_section[section_name]
        if affected_section is None or len(affected_section) == 0:
            # TODO: Consider determining from file what must be present.
            # For now, output a message to assist with refactoring verification.
            print(
                f"The credential for '{cred_def_friendly_name}' not found. Continuing..."
            )
            continue
        password_type = affected_section.get(CONFIG_PASSWORD_TYPE)
        if password_type is None:
            raise CredentialTypeNotFoundError(
                f"restore_keyring_secrets: Cannot find credential type."
            )
        base64_cred_str = affected_section[affected_config_path_parts[-1]]
        cba_password = CredentialByteArray(base64_cred_str.encode("utf-8"))

        desc_cred = DescribedCredential.create_from_base64(
            config_name=storage_def_name,
            credential_name=cred_def.store_credential_name,
            password_type_code=password_type,
            cba_password_base64=cba_password,
        )

        if password_type == CONFIG_PASSWORD_KIND_FILENAME:

            credential = desc_cred.credential
            encrypt_key = False
            if credential.is_password_protected and not credential.is_private_key_ready:
                print(f"An OAuth file name credential is password-protected.")
                print(
                    f"To validate the filename, you need to enter the backup password."
                )
                prompt_for_password_unlock_credential(
                    credential=credential,
                )
                encrypt_key = True
            filename = credential.the_key.decode("utf-8")
            while not os.path.isfile(filename):
                print(f"The credential file does not exist: {filename}")
                user_specified_filename = prompt_user_password_file_does_not_exist(
                    filename
                )
                if user_specified_filename is None:
                    raise CredentialSecretFileNotFoundError(
                        f"An existing credential password filename was not specified, "
                        f"aborting restore."
                    )
                if os.path.isfile(user_specified_filename):
                    credential.set(
                        the_key=CredentialByteArray(
                            user_specified_filename.encode("utf-8")
                        )
                    )
                    filename = user_specified_filename
                    break
                print(
                    f"ERROR: The filename entered does not exist: {user_specified_filename}"
                )
                print(
                    f"You must specify a path to an existing .json password file "
                    f"(i.e., a service account .json OAuth2 file)."
                )
            print(f"The OAuth file exists: {filename}")
            if encrypt_key:
                print(f"Re-encrypting filename credential.")
                credential.encrypt_key()

        CredentialStore().set_credential(desc_cred=desc_cred)
        affected_section[
            affected_config_path_parts[-1]
        ] = CONFIG_KEY_VALUE_KEYRING_INDIRECTION



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""ATBU Configuration-related classes/functions.
- AtbuConfig is the primary class.
"""

import fnmatch
import glob
import logging
import os
from pathlib import Path
import re
from typing import Union
import copy
import json
from uuid import uuid4
import keyring
from send2trash import send2trash

from atbu.mp_pipeline.mp_global import (
    get_verbosity_level,
)
from atbu.common.util_helpers import (
    convert_to_pathlib_path,
    pathlib_splitdrive,
    deep_union_dicts,
    is_absolute_path,
    prompt_YN,
    create_numbered_backup_of_file,
    get_trash_bin_name,
)

from atbu.tools.backup.config_migration_helpers import (
    upgrade_storage_definitions_from_001_to_002,
)

from .constants import *
from .exception import *
from .storage_def_credentials import (
    StorageDefCredentialSet,
    restore_keyring_secrets,
)


def parse_storage_def_specifier(storage_location) -> str:
    what_which = storage_location.lower().split(":", maxsplit=1)
    is_storage_definition_specifier = (
        len(what_which) == 2 # ensure two components
        and len(what_which[0]) > 2 # disambiguate "storage:" specifier from drive letter.
        and what_which[0]
        == CONFIG_SECTION_STORAGE_DEFINITION_SPECIFIER_PREFIX[: len(what_which[0])]
    )
    if is_storage_definition_specifier:
        return what_which[1]
    return None


_STORAGE_DEFINITION_CONFIG_OVERRIDES = {
}

_AUTOMATED_TESTING_MODE = False

def set_automated_testing_mode(enabled: bool):
    global _AUTOMATED_TESTING_MODE
    _AUTOMATED_TESTING_MODE = enabled

def get_automated_testing_mode():
    return _AUTOMATED_TESTING_MODE

def update_config_list_to_dict(items: list):
    """Given the path/name and value pairs in the items list, build a dict
    with all config values represented by items.

    Example::

        given items=[("a.b.c","1"), ("a.e.f","2")]
        return {
            "a": {
                "b": {
                    "c": "1"
                }
                "e": {
                    "f": "2"
                }
            }
        }
    """
    if not isinstance(items, list):
        raise ValueError(f"The items parameter must be a list")
    if not isinstance(items[0], tuple):
        if len(items) % 2 != 0:
            raise ValueError(
                f"Each config location/name must have an accompanying value."
            )
        it = iter(items)
        items = [*zip(it, it)]
    result = {}
    for p in items:
        dest_node = result
        updated_sections = p[0].split(".")
        value = p[1]
        for i, s in enumerate(updated_sections):
            if i >= len(updated_sections) - 1:
                break
            if s not in dest_node:
                dest_node[s] = {}
            dest_node = dest_node[s]
        if s is None:  # pylint: disable=undefined-loop-variable
            raise InvalidStateError(
                f"update_config_list_to_dict: expect 's' to be set "
                f"to one before last period-separated values."
            )
        dest_node[s] = value  # pylint: disable=undefined-loop-variable
    return result


def add_config_items(items: list, config: dict):
    """Merge the dict created from items into config, return the new config to caller."""
    if config is None:
        config = {}
    updates = update_config_list_to_dict(items=items)
    if get_verbosity_level() >= 2:
        logging.debug(f"add_config_items: updates={updates}")
    deep_union_dicts(dest=config, src=updates)
    return config


def add_path_value(
    storage_def_dict: dict, section_value_path: Union[str, list[str]], value: str
):
    if section_value_path is None:
        raise ValueError(f"section_path is required")
    if isinstance(section_value_path, str):
        section_value_path = section_value_path.split("-")
    if not section_value_path or not isinstance(section_value_path, list):
        raise ValueError(
            f"section_path must be in the format section-section-...-value_name"
        )
    value_cfg = storage_def_dict
    for s in section_value_path[:-1]:
        if s not in value_cfg:
            value_cfg[s] = {}
        value_cfg = value_cfg[s]
    value_cfg[section_value_path[-1]] = value


re_not_allowed_cfg_name = re.compile(r"[^a-z0-9_-]")


def is_storage_def_name_ok(storage_def_name):
    if len(storage_def_name) > MAX_STORAGE_DEF_NAME:
        return False
    return not re_not_allowed_cfg_name.search(storage_def_name)


def is_existing_filesystem_storage_path(storage_location: str):
    """An existing valid filesystem storage path has the following characteristics:

    It exists as a directory.

    It is specified as an absolute path. (For no other reason than to ensure the user specifies
        something explicit/clear.)
    """
    return os.path.isdir(storage_location) and is_absolute_path(storage_location)


class AtbuConfigNode:
    def __init__(self, cfg: dict):
        self._cfg = cfg
        if not self._cfg:
            self._cfg = {}

    @property
    def cfg(self):
        return self._cfg

    @cfg.setter
    def cfg(self, value):
        self._cfg = value

    def set_config_from_items(self, items: list):
        """Interpret items as pairs of following...

            ['dot.separated.string', 'string_value', ...]

        ...where the first of the pair, the dot-sep string,
        is interpreted as a path into the dict config, where
        the last component is the value name, and 'string_value'
        its value.

        Note, this is not schema checked and is complete
        free-format so can cause invalid configurations.
        """
        if not isinstance(items, list) or len(items) % 2 != 0:
            raise ValueError(
                f"The items are invalid and must specify a two items, "
                f"a dot-separated path and value, for each affected config item. "
                f"items:{items}"
            )
        # Merge the caller-specified config nodes/values into the existing configuration.
        add_config_items(items, self._cfg)

    def _get_top_section(self, name) -> dict:
        section = self._cfg.get(name)
        if not section:
            self._cfg[name] = {}
            section = self._cfg[name]
        if not isinstance(section, dict):
            raise ConfigSectionNotFound(
                f"The section '{name}' config section is not the expected dictionary."
            )
        return section

    def get_json(self):
        return json.dumps(self._cfg, indent=4)


class AtbuConfig(AtbuConfigNode):

    STARTING_CONFIG = {
        CONFIG_VALUE_NAME_CONFIG_NAME: ATBU_CONFIG_NAME,
        CONFIG_VALUE_NAME_VERSION: ATBU_CONFIG_FILE_VERSION_STRING_CURRENT,
        CONFIG_SECTION_GENERAL: {},
        CONFIG_SECTION_STORAGE_DEFINITIONS: {},
    }

    always_migrate: bool = False

    def __init__(self, path: Union[str, Path], cfg: dict = None):
        super().__init__(cfg=cfg)
        self._path = convert_to_pathlib_path(path)

    @staticmethod
    def _get_new_unique_id():
        return str(uuid4())

    @staticmethod
    def create_starting_config():
        return copy.deepcopy(AtbuConfig.STARTING_CONFIG)

    @staticmethod
    def create_from_file(path: Union[str, Path], create_if_not_exist: bool = False):
        atbu_cfg = AtbuConfig(path=path)
        atbu_cfg.load_config_file(create_if_not_exist=create_if_not_exist)
        return atbu_cfg

    @staticmethod
    def _check_upgrade_default_config() -> None:
        """Trigger upgrade by loading the default <HOME>/<USER>/... configuration.
        This is not used after the <HOME>/<UESR> configuration is upgraded/removed
        and split into individual configuration files.
        """
        default_config_path = AtbuConfig.get_user_default_config_file_path()
        if not default_config_path.is_file():
            return
        cfg = AtbuConfig(path=AtbuConfig.get_user_default_config_file_path())
        cfg.load_config_file(create_if_not_exist=False)
        # Upgrade successful, default config is now seperate per-storage-def config
        # files. Delete default configuration which is no longer needed.
        cfg.delete_config_file()

    @staticmethod
    def access_cloud_storage_config(
        storage_def_name: str,
        must_exist: bool = False,
        create_if_not_exist: bool = False,
        storage_def_dict_not_exist_ok: bool = False,
    ) -> tuple[object, str, dict]:
        AtbuConfig._check_upgrade_default_config()
        config_path = AtbuConfig.get_user_storage_def_config_file_path(
            storage_def_name=storage_def_name,
        )
        if not create_if_not_exist and not config_path.exists():
            if must_exist:
                raise StorageDefinitionNotFoundError(
                    f"The storage definition '{storage_def_name}' not found. "
                    f"The storage config file does not exist: {config_path}",
                )
            return None, None, None
        already_existed = config_path.exists()
        atbu_cfg = AtbuConfig(path=config_path)
        atbu_cfg.load_config_file(create_if_not_exist=create_if_not_exist)
        storage_def_dict = atbu_cfg.get_storage_def_dict(
            storage_def_name=storage_def_name,
            must_exist=False,
        )
        if not storage_def_dict_not_exist_ok and already_existed and storage_def_dict is None:
            # Not newly created so storage_def_dict for the storage_def_name is expected.
            raise InvalidStateError(
                f"The storage config was found but does not contain the storage def config: "
                f"storage_def_name={storage_def_name} "
                f"config_path={config_path}"
            )
        return atbu_cfg, storage_def_name, storage_def_dict

    @staticmethod
    def access_filesystem_storage_config(
        storage_location_path: Union[str, Path],
        resolve_storage_def_secrets: bool = False,
        create_if_not_exist: bool = True,
        prompt_to_create: bool = True,
    ) -> tuple[object, str, dict]:  # object is AtbuConfig
        storage_location_path = convert_to_pathlib_path(storage_location_path)
        storage_atbu_dir: Path = storage_location_path / ATBU_DEFAULT_CONFIG_DIR_NAME
        storage_atbu_cfg_path: Path = storage_atbu_dir / ATBU_USER_DEFAULT_CONFIG_FILE_NAME
        if not storage_atbu_cfg_path.is_file():
            user_says_ok_to_create = "y"
            if prompt_to_create:
                print(
                    f"""
The configuration file does not exist: {storage_atbu_cfg_path}
If you specified the correct location and are aware there should not be an existing
configuration, you can choose to have one created for you.
"""
                )
                user_says_ok_to_create = prompt_YN(
                    prompt_msg=f"Config to create: {storage_atbu_cfg_path}",
                    prompt_question=f"Create config file at this location? ",
                    default_enter_ans="n",
                )
            if not create_if_not_exist or user_says_ok_to_create != "y":
                raise ConfigFileNotFoundError(
                    f"Cannot find the .atbu configuration file: {str(storage_atbu_cfg_path)}"
                )
        storage_atbu_cfg: AtbuConfig = AtbuConfig(path=storage_atbu_cfg_path)
        storage_atbu_cfg.load_config_file(create_if_not_exist=create_if_not_exist)
        storage_def_name, storage_def_dict = storage_atbu_cfg.find_filesystem_storage_def(
            storage_path_to_find=storage_location_path
        )
        if (
            storage_atbu_cfg.is_filesystem_backup_config()
            and storage_def_name is not None
            and storage_def_dict is not None
        ):
            #
            # Existing (not new) filesystem storage definition.
            #

            override_config_path = AtbuConfig.get_storage_def_config_file_path_override(
                storage_def_name=storage_def_name,
            )

            if override_config_path is not None:
                override_atbu_cfg = AtbuConfig(path=override_config_path)
                override_atbu_cfg.load_config_file(create_if_not_exist=False)
                override_atbu_cfg.set_storage_container(
                    container_name=storage_atbu_cfg.get_storage_container()
                )
                if override_atbu_cfg.storage_def_id != storage_atbu_cfg.storage_def_id:
                    raise InvalidConfiguration(
                        f"The storage definition IDs do not match: "
                        f"discovered={storage_atbu_cfg.storage_def_id} "
                        f"override={override_atbu_cfg.storage_def_id} "
                        f"discovered path={storage_atbu_cfg.path} "
                        f"override path={override_atbu_cfg.path}"
                    )
                storage_atbu_cfg = override_atbu_cfg

            # pylint: disable=protected-access
            storage_atbu_cfg._remap_filesystem_storage_backup_info_dirs()

            if resolve_storage_def_secrets:
                storage_def_dict = (
                    storage_atbu_cfg.get_storage_def_with_resolved_secrets_deep_copy(
                        storage_def_name=storage_def_name
                    )
                )

        return storage_atbu_cfg, storage_def_name, storage_def_dict

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = Path(value)

    @property
    def version(self) -> str:
        if not self.cfg:
            raise InvalidConfiguration(f"Cannot find the configuration.")
        if not self.cfg.get(CONFIG_VALUE_NAME_VERSION):
            raise InvalidConfiguration(f"Cannot find the configuration's name.")
        return self.cfg.get(CONFIG_VALUE_NAME_VERSION)

    def create_default_config(self):
        if not self._path:
            raise ValueError(
                "The path is not set, cannot create a default configuration."
            )
        logging.info(f"Writing new configuration: {self._path}")
        self._cfg = AtbuConfig.create_starting_config()
        self.save_config_file()

    def create_config_file_numbered_backup(self):
        backup_path = create_numbered_backup_of_file(
            path=self.path,
            not_exist_ok=True,
        )
        if backup_path is None:
            if self.path.is_file():
                raise AtbuException(f"Unable to backup config file: {str(self.path)}")
            logging.warning(
                f"The config file does not exist so was not backed up: {str(self.path)}"
            )
            return
        logging.info(
            f"The config file was backed up: {str(self.path)} -> to -> {str(backup_path)}"
        )

    def delete_config_file(self):
        bin_name = get_trash_bin_name()
        if self._path.is_file():
            logging.info(f"Removing config file, sending to {bin_name}: {self._path}")
            send2trash(paths=str(self._path))

    def save_config_file(self):
        if self._path.is_file():
            backup_path: Path = self._path.with_suffix(self._path.suffix + ".bak")
            backup_path.unlink(missing_ok=True)
            self._path.rename(backup_path)
        self._path.parent.mkdir(parents=True, exist_ok=True)
        with open(self._path, "w", encoding="utf-8") as config_file:
            config_file.write(json.dumps(self._cfg, indent=4))

    def load_config_file(self, create_if_not_exist: bool = True):
        if self._path.is_dir():
            raise ConfigFileNotFoundError(
                f"The path '{self._path}' must be a file not a directory."
            )
        if not self._path.exists():
            if create_if_not_exist:
                self.create_default_config()
            if not self._path.exists() or not self._path.is_file():
                raise ConfigFileNotFoundError(
                    f"The specified configuration file does "
                    f"not exist or is not a file. {self._path}"
                )
        cfg: dict = None
        try:
            with open(self._path, "r", encoding="utf-8") as cfg_file:
                cfg = json.load(cfg_file)
        except BaseException as err:
            logging.error(
                f"Specified configuration file exists and "
                f"appears to be invalid: {exc_to_string(err)}"
            )
            raise
        if (
            not cfg.get(CONFIG_VALUE_NAME_CONFIG_NAME)
            or cfg[CONFIG_VALUE_NAME_CONFIG_NAME] != ATBU_CONFIG_NAME
        ):
            logging.error(f"The config name is invalid: {self._path}")
            raise InvalidConfigurationFile(f"The config name is invalid: {self._path}")
        version_str = cfg.get(CONFIG_VALUE_NAME_VERSION)
        if version_str not in [
            ATBU_CONFIG_FILE_VERSION_STRING_0_01,
            ATBU_CONFIG_FILE_VERSION_STRING_0_02,
            ATBU_CONFIG_FILE_VERSION_STRING_CURRENT
        ]:
            logging.error(
                f"The config version '{version_str}' is invalid: {self._path}"
            )
            raise InvalidConfigurationFile(
                f"The config version '{version_str}' is invalid: {self._path}"
            )
        if version_str == ATBU_CONFIG_FILE_VERSION_STRING_0_01:
            self._upgrade_version_001_to_002(cfg=cfg)
        version_str = cfg.get(CONFIG_VALUE_NAME_VERSION)
        if version_str == ATBU_CONFIG_FILE_VERSION_STRING_0_02:
            self._upgrade_version_002_to_003(cfg=cfg)
        else:
            self._cfg = cfg

    def _upgrade_version_001_to_002(self, cfg: dict):
        print(f"Configuration file: {self.path}")
        print(
            f"Current configuration file version: {cfg.get(CONFIG_VALUE_NAME_VERSION)}"
        )
        print(f"Required configuration file version: {ATBU_CONFIG_FILE_VERSION_STRING_CURRENT}")
        if self.always_migrate:
            a = "y"
        else:
            a = prompt_YN(
                prompt_msg=(
                    f"You must upgrade your configuration file from "
                    f"{cfg.get(CONFIG_VALUE_NAME_VERSION)} to "
                    f"{ATBU_CONFIG_FILE_VERSION_STRING_CURRENT} before proceeding."
                ),
                prompt_question=f"Proceed with configuration upgrade?",
                default_enter_ans="n",
            )
            if a != "y":
                raise InvalidConfigurationFile(
                    f"The configuration was not upgraded. "
                    f"You cannot use this version of the software with your current configuration. "
                    f"Please use the proper version of software or upgrade your configuration files. "
                    f"Ensure you have a backup of your configuration using the older software before upgrading."
                )
            # After accepting initial migration, accept all others.
            # Example: The first prompt is usually to upgrade the global config
            # file (i.e., c:\Users\SomeUser\.atbu\atbu-config.json). When performing
            # an operation on a local file system backup, that local file system backup
            # configuration should also be upgraded.
            AtbuConfig.always_migrate = True
        self.create_config_file_numbered_backup()
        # Migrate from 0.01 to 0.02.
        upgrade_storage_definitions_from_001_to_002(cfg=cfg)
        # Success
        cfg[CONFIG_VALUE_NAME_VERSION] = ATBU_CONFIG_FILE_VERSION_STRING_0_02
        # Migration successful, save configuration file.
        self._cfg = cfg
        self.save_config_file()

    def _upgrade_version_002_to_003(self, cfg: dict):
        print(f"Configuration file: {self.path}")
        print(f"Current configuration file version: {cfg.get(CONFIG_VALUE_NAME_VERSION)}")
        print(f"Required configuration file version: {ATBU_CONFIG_FILE_VERSION_STRING_0_03}")
        if self.always_migrate:
            a = "y"
        else:
            a = prompt_YN(
                prompt_msg=(
                    f"You must upgrade your configuration file from "
                    f"{cfg.get(CONFIG_VALUE_NAME_VERSION)} to "
                    f"{ATBU_CONFIG_FILE_VERSION_STRING_0_03} before proceeding."
                ),
                prompt_question=f"Proceed with configuration upgrade?",
                default_enter_ans="n",
            )
            if a != "y":
                raise InvalidConfigurationFile(
                    f"The configuration was not upgraded. "
                    f"You cannot use this version of the software with your current configuration. "
                    f"Please use the proper version of software or upgrade your configuration "
                    f"files. Ensure you have a backup of your configuration using the older "
                    f"software before upgrading."
                )
            # After accepting initial migration, accept all others.
            AtbuConfig.always_migrate = True
        self.create_config_file_numbered_backup()
        # Migrate from 0.02 to 0.03.

        #
        # From the configuration, extract individual storage definitions into their own
        # configuration files.
        #
        cfg_dir = self._path.parent
        storage_def_name: str
        storage_def_dict: dict
        storage_definitions = cfg[CONFIG_SECTION_STORAGE_DEFINITIONS]
        is_filesystem = False
        for storage_def_name, storage_def_dict in storage_definitions.items():
            print(f"Processing storage definition: {storage_def_name}...")
            interface_type = storage_def_dict.get(CONFIG_VALUE_NAME_INTERFACE_TYPE)
            is_filesystem = False
            if interface_type == CONFIG_INTERFACE_TYPE_FILESYSTEM:
                is_filesystem = True
            if is_filesystem:
                # Overwrite same file.
                new_cfg_path = self._path
                if len(storage_definitions) != 1:
                    raise InvalidConfiguration(
                        f"Expecting file system config to only have a single storage definition: "
                        f"len(storage_definitions)={len(storage_definitions)}"
                    )
            else:
                # Make new separate file.
                new_cfg_path = (
                    cfg_dir / AtbuConfig.get_user_storage_def_config_file_name(
                        storage_def_name=storage_def_name
                    )
                )
            if new_cfg_path.is_file():
                new_cfg_path_backup = create_numbered_backup_of_file(
                    path=new_cfg_path,
                    not_exist_ok=False,
                )
                print(f"Backed up: {str(new_cfg_path)} --> {str(new_cfg_path_backup)}")
            if new_cfg_path.is_dir():
                raise InvalidStateError(
                    f"The new config path should not be a directory: {str(new_cfg_path)}"
                )
            new_cfg_dict = copy.deepcopy(cfg)
            new_cfg_dict[CONFIG_VALUE_NAME_VERSION] = ATBU_CONFIG_FILE_VERSION_STRING_CURRENT
            new_cfg_gen_sec = new_cfg_dict[CONFIG_SECTION_GENERAL]
            if CONFIG_VALUE_NAME_BACKUP_INFO_DIR in new_cfg_gen_sec:
                # For prior version, CONFIG_VALUE_NAME_BACKUP_INFO_DIR was a str.
                # Going forward, CONFIG_VALUE_NAME_BACKUP_INFO_DIR is an array/list.
                new_cfg_gen_sec[CONFIG_VALUE_NAME_BACKUP_INFO_DIR] = [
                    new_cfg_gen_sec[CONFIG_VALUE_NAME_BACKUP_INFO_DIR]
                ]
            else:
                new_cfg_gen_sec[CONFIG_VALUE_NAME_BACKUP_INFO_DIR] = []
            new_storage_def_dict = copy.deepcopy(storage_def_dict)
            if not new_storage_def_dict.get(CONFIG_VALUE_NAME_STORAGE_DEF_UNIQUE_ID):
                new_storage_def_dict[
                    CONFIG_VALUE_NAME_STORAGE_DEF_UNIQUE_ID
                ] = AtbuConfig._get_new_unique_id()
            new_cfg_dict[CONFIG_SECTION_STORAGE_DEFINITIONS] = {
                storage_def_name: new_storage_def_dict,
            }
            new_cfg = AtbuConfig(
                path=new_cfg_path,
                cfg=new_cfg_dict,
            )
            print(f"Saving {storage_def_name}: {new_cfg.path}")
            new_cfg.save_config_file()
            if is_filesystem:
                break
        if is_filesystem:
            # File system's update without changing the file name so self is
            # the updated file system configuration. Simply reload self.
            self.load_config_file()
        else:
            cfg[CONFIG_SECTION_STORAGE_DEFINITIONS] = {}
            cfg[CONFIG_VALUE_NAME_VERSION] = ATBU_CONFIG_FILE_VERSION_STRING_0_03
            self._cfg = cfg

    def get_general_section(self) -> dict:
        return self._get_top_section(CONFIG_SECTION_GENERAL)

    def get_name(self) -> str:
        if not self.cfg:
            raise InvalidConfiguration(f"Cannot find the configuration.")
        name = self.cfg.get(CONFIG_VALUE_NAME_CONFIG_NAME)
        if not name:
            raise InvalidConfiguration(f"Cannot find the configuration's name.")
        return name

    def validate_name_version(self):
        name = self.get_name()
        version = self.version
        if name != ATBU_CONFIG_NAME:
            raise InvalidConfiguration(
                f"Invalid config name: Expected='{ATBU_CONFIG_NAME}' but got '{name}'"
            )
        if version != ATBU_CONFIG_FILE_VERSION_STRING_CURRENT:
            raise InvalidConfiguration(
                f"Invalid config version: "
                f"Expected='{ATBU_CONFIG_FILE_VERSION_STRING_CURRENT}' but got '{name}'"
            )

    def get_storage_defs_section(self) -> dict:
        return self._get_top_section(CONFIG_SECTION_STORAGE_DEFINITIONS)

    @staticmethod
    def get_user_default_config_dir() -> Path:
        return Path.home() / ATBU_DEFAULT_CONFIG_DIR_NAME

    @staticmethod
    def get_user_default_config_file_path() -> Path:
        return AtbuConfig.get_user_default_config_dir() / ATBU_USER_DEFAULT_CONFIG_FILE_NAME

    @staticmethod
    def get_user_storage_def_config_file_name(storage_def_name: str) -> str:
        return f"{ATBU_STGDEF_CONFIG_FILE_NAME_PREFIX}{storage_def_name}.json"

    @staticmethod
    def get_storage_def_config_file_path_override(
        storage_def_name: str,
    ) -> Path:
        if storage_def_name in _STORAGE_DEFINITION_CONFIG_OVERRIDES:
            return Path(_STORAGE_DEFINITION_CONFIG_OVERRIDES[storage_def_name])
        return None

    @staticmethod
    def get_user_storage_def_config_file_path(
        storage_def_name: str,
        with_overrides: bool = True,
    ) -> Path:
        if with_overrides and storage_def_name in _STORAGE_DEFINITION_CONFIG_OVERRIDES:
            return Path(_STORAGE_DEFINITION_CONFIG_OVERRIDES[storage_def_name])
        return (
            AtbuConfig.get_user_default_config_dir() /
            AtbuConfig.get_user_storage_def_config_file_name(storage_def_name=storage_def_name)
        )

    @staticmethod
    def get_default_backup_info_dir() -> Path:
        """Get the original ATBU default configuration backup information directory.
        This directory was subordinate to the older default configuration file folder.
        It continues to be the default primary backup information directory but can be
        overridden and not used.

        Raises:
            InvalidConfiguration: If the default location cannot be a directory.

        Returns:
            Path: The default backup information directory.
        """
        bid = AtbuConfig.get_user_default_config_dir() / ATBU_DEFAULT_BACKUP_INFO_SUBDIR
        if bid.is_file():
            raise InvalidConfiguration(
                f"Expecting a directory but observed the default backup info directory is a file: ",
                f"{str(bid)}",
            )
        bid.mkdir(parents=True, exist_ok=True)
        return bid

    def get_config_backup_info_dir(self) -> Path:
        """Get the configuration file's backup information directory. The configuration
        backup information directory is the ./backup-info-dir subordinate to the storage
        definition configuration file. This directory may or may not be in the default
        HOME ATBU configuration directory.

        Returns:
            Path: The config backup information directory.
        """
        return self._path.parent / ATBU_DEFAULT_BACKUP_INFO_SUBDIR

    re_extract_storage_def_name = re.compile(
        rf".*\{os.sep}{ATBU_STGDEF_CONFIG_FILE_NAME_PREFIX}([^\s\.]+).json"
    )

    @dataclass
    class StorageDefConfigPathInfo:
        storage_def_name: str
        config_file_path: str

    @staticmethod
    def get_user_storage_def_config_path_info(
        glob_pattern="*"
    ) -> list[StorageDefConfigPathInfo]:
        AtbuConfig._check_upgrade_default_config()
        glob_pattern = (
            AtbuConfig.get_user_default_config_dir() /
            f"{ATBU_STGDEF_CONFIG_FILE_NAME_PREFIX}{str(glob_pattern)}.json"
        )
        remaining_overrides = copy.deepcopy(_STORAGE_DEFINITION_CONFIG_OVERRIDES)
        cpi_list: list[AtbuConfig.StorageDefConfigPathInfo] = []
        for config_file_path in glob.glob(pathname=str(glob_pattern)):
            m = AtbuConfig.re_extract_storage_def_name.match(config_file_path)
            if m is None:
                raise InvalidStateError(
                    f"Cannot extract storage def name from config path: {config_file_path}"
                )
            storage_def_name = m.groups()[0].lower()
            cpi = AtbuConfig.StorageDefConfigPathInfo(
                storage_def_name=storage_def_name,
                config_file_path=config_file_path,
            )
            if cpi.storage_def_name in remaining_overrides:
                cpi.config_file_path = remaining_overrides[cpi.storage_def_name]
                del remaining_overrides[cpi.storage_def_name]
            cpi_list.append(cpi)
        for ovr_def_name, ovr_cfg_path in remaining_overrides.items():
            cpi = AtbuConfig.StorageDefConfigPathInfo(
                storage_def_name=ovr_def_name,
                config_file_path=ovr_cfg_path,
            )
            cpi_list.insert(0, cpi)
        return cpi_list

    @staticmethod
    def get_user_storage_def_names(fnmatch_pattern="*") -> list[str]:
        AtbuConfig._check_upgrade_default_config()
        return [
            cpi.storage_def_name
            for cpi in AtbuConfig.get_user_storage_def_config_path_info()
            if fnmatch.fnmatch(cpi.storage_def_name, fnmatch_pattern)
        ]

    @staticmethod
    def is_user_storage_def_exists(storage_def_name: str) -> bool:
        return len(AtbuConfig.get_user_storage_def_names(fnmatch_pattern=storage_def_name)) != 0

    def get_filesystem_storage_def_root_dir(self) -> Path:
        """Get the file system storage definition's root directory. This method
        assumes the storage definition's container is accurate/resolved.

        Raises:
            InvalidConfiguration: Raised if not a file system storage definition.

        Returns:
            Path: The Path object to the root of the file system storage definition.
        """
        if not self.is_filesystem_backup_config():
            raise InvalidConfiguration(
                f"Configuration is not a filesystem storage definition. "
                f"Cannot get filesystem storage definition root directory: "
                f"path={self.path}"
            )
        return Path(self.get_storage_container())

    def get_filesystem_storage_def_config_dir(self) -> Path:
        return self.get_filesystem_storage_def_root_dir() / ATBU_DEFAULT_CONFIG_DIR_NAME

    def get_filesystem_storage_def_backup_info_dir(self) -> Path:
        return self.get_filesystem_storage_def_config_dir() / ATBU_DEFAULT_BACKUP_INFO_SUBDIR

    @staticmethod
    def _has_replacement_root(path: Path):
        path_str = str(path)
        for rep_field in REPLACEMENT_FIELDS:
            if path_str.startswith(rep_field):
                return True
        return False

    def _remap_filesystem_storage_backup_info_dirs(self):
        """For all backup information directories defined by this configuration,
        remap any such directories to the current file system storage drive.
        """
        if not self.is_filesystem_backup_config():
            return
        fs_sd_root = self.get_filesystem_storage_def_root_dir()
        fs_sd_drive, fs_sd_root_wo_drive = pathlib_splitdrive(fs_sd_root)
        bid_list = self.get_config_backup_info_dirs()
        if not bid_list:
            return
        remapped_bid_list: list[Path] = []
        for bid in bid_list:
            bid_drive, bid_path_wo_drive = pathlib_splitdrive(bid)
            if fs_sd_root_wo_drive not in bid_path_wo_drive.parents:
                remapped_bid_list.append(bid)
            if fs_sd_drive != bid_drive and not AtbuConfig._has_replacement_root(bid_path_wo_drive):
                bid = fs_sd_drive / bid_path_wo_drive
            remapped_bid_list.append(bid)
        self.set_config_backup_info_dirs(remapped_bid_list)

    def find_filesystem_storage_def(
        self, storage_path_to_find: Path
    ) -> tuple[str, dict]:
        if not self.is_filesystem_backup_config():
            return None, None
        storage_def_name, storage_def_dict = self.get_only_storage_def_dict()
        if storage_def_dict is None:
            return None, None
        actual_drive, storage_path_to_find_wo_drive = os.path.splitdrive(storage_path_to_find)
        config_drive, storage_def_path_wo_drive = os.path.splitdrive(
            storage_def_dict[CONFIG_VALUE_NAME_CONTAINER]
        )
        if storage_path_to_find_wo_drive != storage_def_path_wo_drive:
            return None, None
        if actual_drive != config_drive:
            storage_def_dict[CONFIG_VALUE_NAME_CONTAINER] = str(storage_path_to_find)
        return storage_def_name, storage_def_dict

    @staticmethod
    def _get_config_of_type(
        section: dict,
        key_name: str,
        expected_type: type,
        default_value = None,
    ):
        if section is None:
            return default_value
        value = section.get(key_name)
        if value is None:
            return default_value
        if not isinstance(value, expected_type):
            raise InvalidConfiguration(
                f"{key_name} must be a {expected_type} but is "
                f"type={type(value)} value={value}"
            )
        return value

    @staticmethod
    def _get_config_string(
        section: dict,
        key_name: str,
        default_value: str = None,
    ) -> str:
        return AtbuConfig._get_config_of_type(
            section=section,
            key_name=key_name,
            expected_type=str,
            default_value=default_value,
        )

    @staticmethod
    def _get_config_list(
        section: dict,
        key_name: str,
        default_value: list = None,
    ) -> list:
        return AtbuConfig._get_config_of_type(
            section=section,
            key_name=key_name,
            expected_type=list,
            default_value=default_value,
        )

    @staticmethod
    def _get_config_bool(
        section: dict,
        key_name: str,
        default_value: bool = None,
    ) -> bool:
        return AtbuConfig._get_config_of_type(
            section=section,
            key_name=key_name,
            expected_type=bool,
            default_value=default_value,
        )

    @property
    def storage_def_name(self) -> str:
        return self.get_only_storage_def_dict()[0]

    @property
    def storage_def_id(self) -> str:
        _, storage_def_dict = self.get_only_storage_def_dict()
        return storage_def_dict.get(CONFIG_VALUE_NAME_STORAGE_DEF_UNIQUE_ID)

    def get_storage_container(self) -> str:
        _, storage_def_dict = self.get_only_storage_def_dict(not_exist_ok=False)
        return storage_def_dict.get(CONFIG_VALUE_NAME_CONTAINER)

    def set_storage_container(self, container_name):
        _, storage_def_dict = self.get_only_storage_def_dict(not_exist_ok=False)
        storage_def_dict[CONFIG_VALUE_NAME_CONTAINER] = container_name

    def get_interface_type(self) -> str:
        storage_def_name, storage_def_dict = self.get_only_storage_def_dict()
        if storage_def_name is None or storage_def_dict is None:
            return None
        interface_type = storage_def_dict.get(CONFIG_VALUE_NAME_INTERFACE_TYPE)
        if not interface_type:
            raise InvalidConfiguration(
                f"Cannot find '{CONFIG_VALUE_NAME_INTERFACE_TYPE}' in the storage "
                f"definition dictionary."
            )
        return interface_type

    def is_filesystem_backup_config(self):
        return self.get_interface_type() == CONFIG_INTERFACE_TYPE_FILESYSTEM

    def get_config_backup_info_dirs(self) -> list[Path]:
        gen_sec = self.get_general_section()
        if gen_sec is None:
            return []
        if not isinstance(gen_sec, dict):
            raise InvalidConfiguration(
                f"The {CONFIG_SECTION_GENERAL} section should be a dictionary."
            )
        bid_list = AtbuConfig._get_config_list(gen_sec, CONFIG_VALUE_NAME_BACKUP_INFO_DIR)
        if bid_list is None:
            return []
        if not isinstance(bid_list, list):
            raise InvalidConfiguration(
                f"Expected {CONFIG_VALUE_NAME_BACKUP_INFO_DIR} to be a list and not a "
                f"{type(bid_list)}."
            )
        return [Path(bidstr) for bidstr in bid_list]

    def set_config_backup_info_dirs(self, bid_list: list[Union[str,Path]]):
        gen_sec = self.get_general_section()
        if not isinstance(gen_sec, dict):
            raise InvalidConfiguration(
                f"The {CONFIG_SECTION_GENERAL} section should be a dictionary."
            )
        gen_sec[CONFIG_VALUE_NAME_BACKUP_INFO_DIR] = [str(bid) for bid in bid_list]

    def get_resolved_config_backup_info_dirs(self) -> list[Path]:
        cfg_bids_resolved: list[Path] = []
        cfg_bids = self.get_config_backup_info_dirs()
        for cfg_bid in cfg_bids:
            cfg_bid_str = str(cfg_bid)
            cfg_bid_str = cfg_bid_str.replace(
                REPLACEMENT_FIELD_DEFAULT_CONFIG_DIR,
                str(AtbuConfig.get_user_default_config_dir()),
            )
            cfg_bid_str = cfg_bid_str.replace(
                REPLACEMENT_FIELD_CONFIG_DIR,
                str(self._path.parent),
            )
            cfg_bid = Path(cfg_bid_str).resolve()
            cfg_bids_resolved.append(cfg_bid)
        return cfg_bids_resolved

    def get_backup_info_dirs(self) -> list[Path]:
        """Get the backup information directories for this configuration.
        This will return a list of Path objects, each of which is a backup
        information directory.

        The first Path in the returned list is the "primary" backup information
        directory, and is the one expected to contain the latest backup information,
        where the other Path objects (other than the first one) are considered
        secondary and therefore backup copies of the primary backup information.


        Raises:
            BackupInformationDirectoryNotFound: Raised if a backup information
                directory Path cannot be a directory (i.e., storage medium is
                offline or the Path is to a file), or if no backup information
                directories could be deduced by the user's configuration. There
                must always be at least one (primary) backup information directory.

        Returns:
            list[Path]: The list of backup information directories, the first of
                which is considered the "primary" backup information directory.
        """
        default_bid = AtbuConfig.get_default_backup_info_dir().resolve()
        cfg_bids_resolved = self.get_resolved_config_backup_info_dirs()
        cfg_no_default_bid = AtbuConfig._get_config_bool(
            self.get_general_section(),
            CONFIG_VALUE_NAME_BACKUP_INFO_DIR_NO_DEFAULTS,
            False,
        )
        if not cfg_no_default_bid:
            # Original and still the current default behavior, if not overridden using the
            # CONFIG_VALUE_NAME_NO_DEFAULT_BACKUP_INFO_DIR setting, is for the user's HOME
            # .atbu/backup-info-dir to be the primary backup info dir.
            cfg_bids_resolved.insert(0, default_bid)
            # ...and for filesystem storage definitions, the secondary backup information
            # directory is in the target storage configuration directory (aka the "config"
            # backup information directory).
            if self.is_filesystem_backup_config():
                cfg_bids_resolved.insert(1, self.get_filesystem_storage_def_backup_info_dir())
        cfg_bid_resolved_str_set: set[str] = set()
        cfg_bids_final: list[Path] = []
        for cfg_bid in cfg_bids_resolved:
            if not str(cfg_bid) in cfg_bid_resolved_str_set:
                cfg_bids_final.append(cfg_bid)
                cfg_bid_resolved_str_set.add(str(cfg_bid))
        for cfg_bid in cfg_bids_final:
            if not cfg_bid.exists():
                cfg_bid.mkdir(parents=True, exist_ok=True)
            if not cfg_bid.is_dir():
                raise BackupInformationDirectoryNotFound(
                    f"The backup information directory was not found: {str(cfg_bid)}"
                )
        if not cfg_bids_final:
            raise BackupInformationDirectoryNotFound(
                f"No backup information directories were found. "
                f"There must be at least one backup information directory. "
                f"Your configuration is invalid. "
                f"Please review your configuration's backup information directory settings."
            )
        return cfg_bids_final

    def get_primary_backup_info_dir(self) -> Path:
        return self.get_backup_info_dirs()[0]

    def get_primary_backup_info_file_paths(
        self,
        storage_def_name: str,
    ) -> list[str]:
        pattern = self.get_primary_backup_info_dir() / f"{storage_def_name}*"
        return glob.glob(pathname=str(pattern))

    @staticmethod
    def resolve_storage_location(
        storage_location,
        resolve_storage_def_secrets=False,
        create_if_not_exist=False,
    ) -> tuple[object, str, dict]:  # object is AtbuConfig
        """Given a user-specified storage_location, try to resolve
        to an existing or newly created storage_def.

            Possible return values (always a tuple of 3):
                1)  (self, storage_def_name, storage_def):
                        storage_location is storage def specifier
                        defined in default ATBU configuration.
                2)  (storage_atbu_cfg, storage_def_name, storage_def):
                        storage_location is a file system storage whose
                3)  (None, None, None):
                        storage_location is not resolved to anything.
        """
        # Parse as potential specifier.
        storage_def_name = parse_storage_def_specifier(
            storage_location=storage_location
        )

        storage_atbu_cfg: AtbuConfig = None

        if storage_def_name:
            # Specifier found, if config exists, return it directly or with secrets resolved.
            storage_atbu_cfg, _, storage_def_dict = AtbuConfig.access_cloud_storage_config(
                storage_def_name=storage_def_name,
                must_exist=False,
                create_if_not_exist=False,
            )

            if (
                storage_atbu_cfg is not None
                and storage_def_dict is not None
                and resolve_storage_def_secrets
            ):
                storage_def_dict = storage_atbu_cfg.get_storage_def_with_resolved_secrets_deep_copy(
                    storage_def_name=storage_def_name,
                )

            return storage_atbu_cfg, storage_def_name, storage_def_dict

        # Fall through if storage_def specifier not found.
        # See if storage_location is a filesystem storage.

        if is_existing_filesystem_storage_path(storage_location=storage_location):
            try:
                # storage_location is a directory specified as absolute path.
                (
                    storage_atbu_cfg,
                    storage_def_name,
                    storage_def,
                ) = AtbuConfig.access_filesystem_storage_config(
                    storage_location_path=storage_location,
                    resolve_storage_def_secrets=resolve_storage_def_secrets,
                    create_if_not_exist=create_if_not_exist,
                    prompt_to_create=True,
                )
                return storage_atbu_cfg, storage_def_name, storage_def
            except ConfigFileNotFoundError:
                pass
        # storage_location is not resolved.
        return None, None, None

    def get_storage_def_dict_deep_copy(self, storage_def_name: str) -> dict:
        # TODO search for usage re atbu_cfg
        storage_definitions = self.get_storage_defs_section()
        storage_def = storage_definitions.get(storage_def_name)
        if storage_def is not None:
            storage_def = copy.deepcopy(storage_def)
        return storage_def

    def get_storage_def_with_resolved_secrets_deep_copy(
        self, storage_def_name: str, keep_secrets_base64_encoded: bool = False
    ) -> dict:
        storage_def_dict = self.get_storage_def_dict_deep_copy(
            storage_def_name=storage_def_name,
        )
        if not isinstance(storage_def_dict, dict):
            raise StorageDefinitionNotFoundError(
                f"Cannot find storage definition "
                f"{storage_def_name} in config file {str(self.path)}"
            )

        cred_set = StorageDefCredentialSet(
            storage_def_name=storage_def_name,
            storage_def_dict=storage_def_dict,
        )
        cred_set.populate()
        cred_set.unprotect(
            base64_encoded_secrets=keep_secrets_base64_encoded,
        )
        return cred_set.storage_def_dict

    def get_storage_def_dict(self, storage_def_name, must_exist: bool = False) -> dict:
        storage_def_dict = self.get_storage_defs_section().get(storage_def_name)
        if must_exist and storage_def_dict is None:
            raise StorageDefinitionNotFoundError(
                f"The storage definition '{storage_def_name}' not found."
            )
        return storage_def_dict

    def get_only_storage_def_dict(self, not_exist_ok: bool = True) -> tuple[str, dict]:
        storage_def_section: dict = self.get_storage_defs_section()
        if len(storage_def_section) != 1:
            if not_exist_ok:
                return None, None
            raise InvalidConfiguration(
                f"One storage definition was not found: "
                f"len(storage_def_section)={len(storage_def_section)}"
            )
        storage_def_name, storage_def_dict = list(storage_def_section.items())[0]
        return storage_def_name, storage_def_dict

    def verify_only_one_storage_def(self) -> None:
        storage_def_section: dict = self.get_storage_defs_section()
        if len(storage_def_section) != 1:
            raise InvalidStorageDefinitionFile(
                f"Expected one storage definition configuration but observed :"
                f"{len(storage_def_section)}."
            )

    def is_storage_def_exists(self, storage_def_name) -> bool:
        return self.get_storage_def_dict(storage_def_name=storage_def_name) is not None

    def get_storage_def_compression_section(self, storage_def_name) -> dict:
        storage_def_dict = self.get_storage_def_dict(
            storage_def_name=storage_def_name,
            must_exist=True,
        )
        return storage_def_dict.get(CONFIG_SECTION_COMPRESSION)

    def get_compression_settings_deep_copy(self, storage_def_name) -> dict:
        compression_sect = self.get_storage_def_compression_section(
            storage_def_name=storage_def_name,
        )
        if compression_sect is not None:
            compression_sect = copy.deepcopy(compression_sect)
        else:
            compression_sect = {}
        if not isinstance(compression_sect, dict):
            raise ConfigSectionNotFound(
                f"For '{storage_def_name}' compression section, "
                f"expected dict but got {type(compression_sect)}"
            )
        for k, v in ATBU_BACKUP_COMPRESSION_DEFAULTS.items():
            if k not in compression_sect:
                compression_sect[k] = v
        return compression_sect

    def rename_storage_def(self, old_name, new_name):
        """Rename an existing storage def in this configuration."""
        storage_def_section = self.get_storage_defs_section()
        if old_name not in storage_def_section:
            raise ValueError(
                f"Expected to find {old_name} in the configuration."
            )
        storage_def = storage_def_section.pop(
            old_name
        )
        storage_def_section[new_name] = storage_def

    def get_storage_def_encryption_section(self, storage_def_name) -> dict:
        storage_def = self.get_storage_def_dict(storage_def_name=storage_def_name)
        if not storage_def:
            return None
        return storage_def.get(CONFIG_SECTION_ENCRYPTION)

    def is_storage_def_configured_for_encryption(self, storage_def_name):
        storage_def = self.get_storage_def_dict(storage_def_name=storage_def_name)
        if not storage_def:
            return False
        enc_section: dict = storage_def.get(CONFIG_SECTION_ENCRYPTION)
        if enc_section is None:
            return False
        # The encryption section may exist before any encryption is enabled,
        # but if the "key" value is set, it means encryption has been enabled.
        if enc_section.get(CONFIG_VALUE_NAME_ENCRYPTION_KEY) is None:
            # Not enabled.
            return False
        # Enabled.
        return True

    def delete_storage_def_secrets(self, storage_def_name: str):
        try:
            keyring.delete_password(
                service_name=storage_def_name,
                username=CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD,
            )
        except keyring.errors.PasswordDeleteError:
            pass

        try:
            keyring.delete_password(
                service_name=storage_def_name,
                username=CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
            )
        except keyring.errors.PasswordDeleteError:
            pass

    def delete_storage_def(self, storage_def_name: str):
        if not self.is_storage_def_exists(storage_def_name=storage_def_name):
            return
        self.delete_storage_def_secrets(storage_def_name=storage_def_name)
        del self.get_storage_defs_section()[storage_def_name]

    @staticmethod
    def create_cloud_storage_def_config(
        storage_def_name: str,
    ) -> object:
        if AtbuConfig.is_user_storage_def_exists(storage_def_name=storage_def_name):
            raise StorageDefinitionAlreadyExists(
                f"The backup storage definition '{storage_def_name}' already exists."
            )
        config_path = AtbuConfig.get_user_storage_def_config_file_path(
            storage_def_name=storage_def_name,
        )
        atbu_cfg = AtbuConfig(
            path=config_path,
        )

        atbu_cfg.cfg = AtbuConfig.create_starting_config()

        return atbu_cfg

    def create_storage_def(
        self,
        interface_type: CONFIG_INTERFACE_HINT,
        provider_id: str,
        container: str,
        project: str = None,
        storage_key: str = None,
        storage_secret: str = None,
        other_driver_kv_pairs: dict = None,
        unique_storage_def_name: str = None,
        include_iv: bool = True,
        include_backup_info: bool = True,
    ) -> tuple[str, dict]:
        if unique_storage_def_name is None:
            raise InvalidStorageDefinitionName(
                f"unique_storage_def_name not specified, cannot be None."
            )
        unique_storage_def_name = unique_storage_def_name.lower()
        if not is_storage_def_name_ok(storage_def_name=unique_storage_def_name):
            raise InvalidStorageDefinitionName(
                f"unique_storage_def_name '{unique_storage_def_name}' is invalid."
            )
        if self.is_storage_def_exists(unique_storage_def_name):
            raise StorageDefinitionAlreadyExists(
                f"The backup storage definition '{unique_storage_def_name}' already exists."
            )
        self.get_storage_defs_section()[unique_storage_def_name] = {
            CONFIG_VALUE_NAME_INTERFACE_TYPE: interface_type,
            CONFIG_VALUE_NAME_PROVIDER: provider_id,
            CONFIG_VALUE_NAME_CONTAINER: str(container),
            CONFIG_SECTION_DRIVER: {},
        }
        storage_def_dict = self.get_storage_defs_section()[unique_storage_def_name]
        storage_def_dict[
            CONFIG_VALUE_NAME_STORAGE_DEF_UNIQUE_ID
        ] = AtbuConfig._get_new_unique_id()
        driver_section = storage_def_dict[CONFIG_SECTION_DRIVER]
        if storage_key:
            driver_section[CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY] = storage_key
        if storage_secret:
            driver_section[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET] = storage_secret
        if project:
            driver_section[CONFIG_VALUE_NAME_DRIVER_STORAGE_PROJECT] = project
        if other_driver_kv_pairs:
            for k, v in other_driver_kv_pairs.items():
                driver_section[k] = v
        if include_iv != CONFIG_STORAGE_PERSISTED_IV_DEFAULT_VALUE:
            # Set include_iv since it is not the default value.
            storage_def_dict[CONFIG_SECTION_ENCRYPTION] = {
                CONFIG_VALUE_NAME_STORAGE_PERSISTED_IV: include_iv
            }
        if include_backup_info != CONFIG_STORAGE_PERSISTED_BACKUP_INFO_DEFAULT_VALUE:
            # Set value indicating whether or not to store backup information
            # to the file/cloud storage.
            storage_def_dict[
                CONFIG_VALUE_NAME_STORAGE_PERSISTED_BACKUP_INFO
            ] = include_backup_info
        return (
            unique_storage_def_name,
            storage_def_dict,
        )

    def create_filesystem_storage_def(
        self, storage_location_path: Path, unique_storage_def_name: str = None
    ) -> tuple[str, dict]:
        return self.create_storage_def(
            interface_type=CONFIG_INTERFACE_TYPE_FILESYSTEM,
            provider_id=CONFIG_INTERFACE_TYPE_FILESYSTEM,
            container=str(storage_location_path),
            unique_storage_def_name=unique_storage_def_name,
        )

    def save_storage_def_with_cleartext_secrets(
        self, storage_def_name: Union[str, list], backup_file_path: str
    ):
        if isinstance(storage_def_name, str):
            storage_def_name = [storage_def_name]
        if not isinstance(storage_def_name, list) or len(storage_def_name) == 0:
            raise ValueError(f"No storage definition names specified to restore.")
        # Get copy of current configuration, less storage definitions.
        cfg_for_save = copy.deepcopy(self._cfg)
        storage_definitions_section = {}
        cfg_for_save[CONFIG_SECTION_STORAGE_DEFINITIONS] = storage_definitions_section
        for name in storage_def_name:
            logging.info(f"Getting storage definition {name}...")
            storage_def_copy = self.get_storage_def_with_resolved_secrets_deep_copy(
                storage_def_name=name, keep_secrets_base64_encoded=True
            )
            storage_definitions_section[name] = storage_def_copy
        logging.info(f"Saving backup to {backup_file_path} ...")
        with open(backup_file_path, "w", encoding="utf-8") as backup_file:
            backup_file.write(json.dumps(cfg_for_save, indent=4))
        logging.info(f"Backup complete.")

    def restore_storage_def(
        self,
        storage_def_new_name: str,
        backup_file_path: str,
        prompt_if_exists: bool = True,
    ) -> bool:
        print(f"Loading backup file {backup_file_path}...")
        atbu_stg_cfg = AtbuConfig.create_from_file(path=backup_file_path)
        atbu_stg_cfg.validate_name_version()
        (
            storage_def_old_name,
            storage_def_dict,
        ) = atbu_stg_cfg.get_only_storage_def_dict()
        if storage_def_new_name is None:
            storage_def_new_name = storage_def_old_name
        atbu_stg_cfg.rename_storage_def(
            old_name=storage_def_old_name,
            new_name=storage_def_new_name,
        )
        storage_def_name, storage_def_dict = atbu_stg_cfg.get_only_storage_def_dict()
        if storage_def_name != storage_def_new_name:
            raise InvalidStateError(
                f"Expected storage_def to have new name "
                f"'{storage_def_new_name}' but observed '{storage_def_name}'."
            )
        if storage_def_new_name in self.get_storage_defs_section():
            if prompt_if_exists:
                print(
                    f"""
*** WARNING *** WARNING *** WARNING *** WARNING *** WARNING *** WARNING
The storage definition '{storage_def_new_name}' exists. You are about to
replace it with '{storage_def_old_name}'. If this is an encrypted backup
where the private key is not backed up, you will lose access to all data
in this backup if you delete this configuration.
            """
                )
                a = prompt_YN(
                    prompt_msg=f"You are about to overwrite a backup storage definition.",
                    prompt_question=(
                        f"Are you certain you want to overwrite '{storage_def_new_name}' "
                    ),
                    default_enter_ans="n",
                )
                if a != "y":
                    print("The storage definition will not be overwritten.")
                    return None
        logging.info(
            f"Restoring {storage_def_old_name} as {storage_def_new_name} from {backup_file_path}"
        )
        logging.info(f"Restoring secrets from backup file to keyring.")
        restore_keyring_secrets(
            storage_def_name=storage_def_name,
            storage_def=storage_def_dict,
        )
        # Merge 'general' and 'storage-definitions' sections from import to dest config.
        deep_union_dicts(
            dest=self.get_storage_defs_section(),
            src=atbu_stg_cfg.get_storage_defs_section(),
        )
        deep_union_dicts(
            dest=self.get_general_section(), src=atbu_stg_cfg.get_general_section()
        )
        logging.info(f"Saving configuration {str(self._path)}...")
        self.save_config_file()
        logging.info(f"Configuration updated... restore complete")
        return storage_def_name


def get_atbu_config_from_file(
    config_file_path: str,
) -> tuple[AtbuConfig, str, dict]:
    atbu_cfg = AtbuConfig.create_from_file(
        path=config_file_path,
        create_if_not_exist=False,
    )
    atbu_cfg.verify_only_one_storage_def()
    storage_def_name, storage_def_dict = atbu_cfg.get_only_storage_def_dict()
    return atbu_cfg, storage_def_name, storage_def_dict


def register_storage_def_config_override(
    storage_def_config_path: str,
    only_if_not_already_present: bool,
):
    if not os.path.isfile(storage_def_config_path):
        raise InvalidStorageDefinitionFile(
            f"register_storage_def_config_override: "
            f"Cannot find the config file: {storage_def_config_path}"
        )

    storage_def_name: str = None
    try:
        _, storage_def_name, _ = get_atbu_config_from_file(
            config_file_path=storage_def_config_path,
        )
        if not storage_def_name:
            raise InvalidStorageDefinitionName(
                f"Storage definition name is invalid after reading from config: "
                f"{storage_def_config_path}"
            )
    except ConfigFileNotFoundError as ex:
        logging.error(
            f"The configuration file was not found: {storage_def_config_path}"
        )
        raise

    if (
        only_if_not_already_present
        and AtbuConfig.get_user_storage_def_config_file_path(
            storage_def_name=storage_def_name,
        ).is_file()
    ):
        return
    _STORAGE_DEFINITION_CONFIG_OVERRIDES[storage_def_name] = storage_def_config_path


def unregister_storage_def_config_override(
    storage_def_name: str,
):
    if storage_def_name in _STORAGE_DEFINITION_CONFIG_OVERRIDES:
        del _STORAGE_DEFINITION_CONFIG_OVERRIDES[storage_def_name]



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Restore files.
"""

from concurrent import futures
from concurrent.futures import FIRST_COMPLETED
import glob
import gzip
import io
import os
import logging
from pathlib import Path
import re
import shutil
import tempfile
from typing import Iterator

from atbu.mp_pipeline.mp_global import (
    get_process_pool_exec_init_func,
    get_process_pool_exec_init_args,
)
from atbu.mp_pipeline.mp_pipeline import (
    MultiprocessingPipeline,
    SubprocessPipelineStage,
)
from atbu.mp_pipeline.mp_helper import wait_futures_to_regulate

from .constants import (
    ATBU_FILE_BACKUP_EXTENSION,
    ATBU_FILE_BACKUP_EXTENSION_ENCRYPTED,
)
from .exception import *
from .chunk_reader import ChunkSizeFileReader
from .config import (
    AtbuConfig,
    is_existing_filesystem_storage_path,
    parse_storage_def_specifier,
)
from .storage_interface.base import DEFAULT_MAX_SIMULTANEOUS_FILE_BACKUPS
from .backup_core import (
    BACKUP_OPERATION_NAME_RESTORE,
    BackupAnomaly,
    log_anomalies_report,
    StorageDefinition,
    BackupFileInformation,
    StorageFileRetriever,
    file_operation_futures_to_results,
    BackupPipelineWorkItem,
    is_qualified_for_operation,
    run_operation_stage,
)
from .backup_selections import (
    ensure_glob_pattern_for_dir,
    get_all_specific_backup_file_info,
    verify_specific_backup_selection_list,
    user_specifiers_to_selections,
)
from .global_hasher import GlobalHasherDefinitions


class RestoreFile(StorageFileRetriever):
    def __init__(
        self,
        file_info: BackupFileInformation,
        dest_root_location: str,
        allow_overwrite: bool,
        storage_def: StorageDefinition,
        temp_dir: str,
    ):
        super().__init__(file_info=file_info, storage_def=storage_def)
        self._temp_dir = temp_dir
        self._dest_root_location = Path(dest_root_location)
        self._allow_overwrite = allow_overwrite
        self._dest_path_intermediate = (
            self._dest_root_location / self.file_info.restore_path
        )
        self._dest_path_final = self._dest_path_intermediate
        self._dest_path_for_logging = str(self._dest_path_intermediate)
        self._dest_path_existed: bool = (
            None  # For sanity check against allow_overwrite.
        )
        self._dest_file = None

    def get_download_iterator(self) -> tuple[Iterator[bytes], tuple[Exception]]:
        if not self.file_info.is_decrypt_operation:
            return super().get_download_iterator()
        iterator = ChunkSizeFileReader(
            path=self.file_info.path,
            fileobj=None,
            read_without_size=False,
            chunk_size=self._storage_def.download_chunk_size,
            user_func=None,
        )
        iterator.open()
        retry_exception_types = ()
        return (
            iterator,
            retry_exception_types,
        )

    def prepare_destination(self):

        if self.file_info.is_decrypt_operation:
            self._dest_path_intermediate = (
                self._dest_root_location / self.preamble_path_without_root
            )
            self._dest_path_final = self._dest_path_intermediate
            self._dest_path_for_logging = str(self._dest_path_intermediate)

        if self.is_compressed:
            #
            # Decomrpession case: decompress to temp file.
            #
            dest_file_fd, self._dest_path_intermediate = tempfile.mkstemp(
                prefix="atbu_z_", dir=self._temp_dir, text=False
            )
            self._dest_path_intermediate = Path(self._dest_path_intermediate)
            self._dest_file = io.FileIO(file=dest_file_fd, mode="w+b", closefd=True)
            # The _dest_path never existed prevously in this case,
            # all cleanup on failure is welcome.
            self._dest_path_existed = False
            # Disable default cleartext hashing because completion code will
            # perform a cleartext hashing on the final decompressed file.
            self.disable_cleartext_hashing()
            return

        #
        # This path is the non-decompression case:
        #   _dest_path == _dest_path_final.
        #

        #
        # Sample the existence of the destination path before any
        # file operation are performed. For a case where the code
        # further below raises an exception due to an existing file
        # and overwrite not allowed, we do not want any cleanup code
        # touching what would be some file that existed prior to
        # any of our efforts.
        #
        if self._dest_path_existed is None:
            # Sampled once, this is the existence state before restore efforts.
            self._dest_path_existed = self._dest_path_intermediate.exists()

        #
        # Do not allow restore if destination path already exists and
        # overwrite is not allowed.
        #
        if self._dest_path_intermediate.exists() and not self._allow_overwrite:
            # In case it's unclear, regarding retries:
            # The 'attempt_failed_cleanup' method cleans up self._dest_path
            # on failure, so the following should pass, even on a retry attempt,
            # assuming there was no file there prior to this restore.
            raise RestoreFilePathAlreadyExistsError(
                f"The restore file destination path already exists and "
                f"allow_overwrite={self._allow_overwrite}: {str(self._dest_path_intermediate)}"
            )
        self._dest_path_intermediate.parent.mkdir(parents=True, exist_ok=True)
        self._dest_file = open(str(self._dest_path_intermediate), "wb")

    def attempt_failed_cleanup(self):
        """Called after failed attempt. State at this point
        is either allow_overwrite==True or the dest file did
        exist to begin with. Either way, cleanup for next
        retry attempt.
        """
        if self._dest_file:
            self._dest_file.close()
            self._dest_file = None
        if self._dest_path_existed is None:
            # Must be set by prepare.
            raise InvalidStateError(
                f"The self.dest_path_existed flag " f"should not be None at this point."
            )

        # Only allow cleanup of _dest_path if overwrite was
        # allowed with an existing file, or if no existing
        # file was present regardless of whether or not
        # overwrite was allowed (because in this latter
        # case no file existed beforehand).
        if self._dest_path_intermediate.exists() and (
            self._allow_overwrite or not self._dest_path_existed
        ):
            logging.info(
                f"Deleting file as part of retry prep: "
                f"allow_overwrite={self._allow_overwrite} "
                f"dest_path_existed={self._dest_path_existed} "
                f"path={self._dest_path_for_logging}"
            )
            # Delete file after a failed attempt only if either overwrite
            # is allowed or the file was not there to begin with.
            os.remove(path=self._dest_path_intermediate)

    def report_progress(self, percent: int):
        logging.info(f"{percent: >3}% completed of {self._dest_path_for_logging}")

    def process_decrypted_chunk(self, decrypted_chunk: bytes):
        if not self._dest_file:
            raise InvalidStateError(
                f"RestoreFile: Expected self.dest_file to be set/open."
            )
        self._dest_file.write(decrypted_chunk)

    def download_completed(self):

        if self._dest_path_intermediate == self._dest_path_final:
            if self._dest_file:
                self._dest_file.close()
                self._dest_file = None
        else:
            if self._dest_path_final.exists() and not self._allow_overwrite:
                # Given earlier checks, this point should never be reached.
                raise RestoreFilePathAlreadyExistsError(
                    f"The restore final file destination path already exists "
                    f"and allow_overwrite={self._allow_overwrite}: "
                    f"{str(self._dest_path_final)}"
                )

            if not self.is_compressed:
                raise InvalidStateError(
                    f"Differing download/final dest paths is only expected "
                    f"for the decompression case. dest_path={self._dest_path_intermediate} "
                    f"final_dest_path={self._dest_path_final}"
                )

            logging.debug(f"Decompressing to {self._dest_path_final}")
            self._dest_path_final.parent.mkdir(parents=True, exist_ok=True)
            self._dest_file.seek(0, io.SEEK_SET)
            with (
                gzip.GzipFile(mode="rb", fileobj=self._dest_file) as input_file_gz,
                open(self._dest_path_final, mode="wb") as output_file,
            ):
                hasher = GlobalHasherDefinitions().create_hasher()
                while True:
                    b = input_file_gz.read(1024 * 1024 * 25)
                    if len(b) == 0:
                        break
                    hasher.update_all(b)
                    output_file.write(b)
                self._cleartext_digest = hasher.get_primary_hexdigest()

            self._dest_file.close()
            self._dest_file = None

            self.total_cleartext_bytes = os.path.getsize(self._dest_path_final)
            os.remove(self._dest_path_intermediate)

        os.utime(
            path=self._dest_path_final,
            times=(
                self.file_info.accessed_time_posix,
                self.file_info.modified_time_posix,
            ),
        )

        self.perform_common_checks(
            log_msg_prefix_str="RestoreFile",
            file_path_for_logging=self._dest_path_for_logging,
            orig_file_info=self.file_info,
        )

    def download_failed(self):
        if self._dest_file:
            self._dest_file.close()
            self._dest_file = None
            try:
                os.remove(path=self._dest_path_intermediate)
            except FileNotFoundError:
                pass

    def final_cleanup(self):
        if self._dest_file:
            self._dest_file.close()
            self._dest_file = None


class Restore:
    MAX_SIMULTANEOUS_FILES = DEFAULT_MAX_SIMULTANEOUS_FILE_BACKUPS

    def __init__(
        self,
        storage_def: StorageDefinition,
        selections: list[BackupFileInformation],
        dest_root_location: str,
        allow_overwrite: bool,
        auto_path_mapping: bool,
    ) -> None:
        if not isinstance(storage_def, StorageDefinition):
            raise ValueError(f"The storage_def must be a StorageDefinition.")
        self._storage_def = storage_def
        if len(selections) == 0:
            raise InvalidStateError(
                f"RestoreFile requires sbs_list to have at "
                f"least one SpecificBackupSelection instance."
            )
        self._selected_files: list[BackupFileInformation] = selections
        self.dest_root_location = dest_root_location
        Path(self.dest_root_location).mkdir(parents=True, exist_ok=True)
        self._temp_dir = tempfile.mkdtemp(
            prefix="atbu_resttmp_", dir=self.dest_root_location
        )
        self._allow_overwrite = allow_overwrite
        self._auto_path_mapping = auto_path_mapping
        self._subprocess_pipeline = MultiprocessingPipeline(
            name="Restore",
            max_simultaneous_work_items=min(os.cpu_count() // 2, 15),
            stages=[
                SubprocessPipelineStage(
                    fn_determiner=is_qualified_for_operation,
                    fn_worker=run_operation_stage,
                )
            ],
            process_initfunc=get_process_pool_exec_init_func(),
            process_initargs=get_process_pool_exec_init_args(),
        )
        self.anomalies: list[BackupAnomaly] = []
        self.success_count = 0
        self._is_used = False

    def shutdown(self):
        if self._subprocess_pipeline is not None:
            self._subprocess_pipeline.shutdown()
        try:

            def on_error(_, path, exc_info):
                value = ""
                if (
                    exc_info is not None
                    and len(exc_info) >= 3
                    and exc_info[2] is not None
                ):
                    value = f" ({exc_info[2]})"
                logging.warning(f"Failed to delete temp file: {path}{value}")

            if (
                self._temp_dir is not None
                and len(self._temp_dir) > 0
                and os.path.isdir(self._temp_dir)
            ):
                shutil.rmtree(
                    path=self._temp_dir, ignore_errors=False, onerror=on_error
                )
        except Exception as ex:
            logging.error(
                f"Unhandled exception while cleaning up "
                f"the restore temp folder: {self._temp_dir} ex={ex}"
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()
        return False

    @property
    def storage_def(self) -> StorageDefinition:
        return self._storage_def

    @property
    def is_failed(self):
        return len(self.anomalies) > 0

    def _restore_files(self):
        if self._is_used:
            raise AlreadyUsedError(f"This instance has already been used.")
        self._is_used = True
        logging.info(f"Starting restore from '{self.storage_def.storage_def_name}'...")
        pending_restores = set()
        try:
            logging.info(f"Scheduling restore jobs...")
            for file_info in self._selected_files:
                if file_info.is_unchanged_since_last:
                    logging.debug(
                        f"unchanged: path={file_info.path} "
                        f"path_wo_root={file_info.path_without_root} "
                        f"backing={file_info.backing_fi.path}"
                    )
                else:
                    logging.debug(
                        f"changed: path={file_info.path} "
                        f"path_wo_root={file_info.path_without_root} "
                        f"backing={file_info.backing_fi is not None}"
                    )
                logging.debug(f"")
                restore_file = RestoreFile(
                    file_info=file_info,
                    dest_root_location=self.dest_root_location,
                    allow_overwrite=self._allow_overwrite,
                    storage_def=self.storage_def,
                    temp_dir=self._temp_dir,
                )
                wait_futures_to_regulate(
                    fs=pending_restores,
                    max_allowed_pending=Restore.MAX_SIMULTANEOUS_FILES,
                )
                future = self._subprocess_pipeline.submit(
                    work_item=BackupPipelineWorkItem(
                        operation_name=BACKUP_OPERATION_NAME_RESTORE,
                        file_info=file_info,
                        is_qualified=True,
                        operation_runner=restore_file,
                    )
                )
                pending_restores.add(future)
        except Exception as ex:
            logging.error(
                f"Unexpected exception during _restore_files. {exc_to_string(ex)}"
            )
            raise
        finally:
            logging.info(f"Wait for restore file operations to complete...")
            while len(pending_restores) > 0:
                done, pending_restores = futures.wait(
                    fs=pending_restores, return_when=FIRST_COMPLETED
                )
                num_completed = len(done)
                fi_complete = []
                new_anomalies: list[BackupAnomaly] = []
                file_operation_futures_to_results(
                    fs=done,
                    fi_list=fi_complete,
                    anomalies=new_anomalies,
                    the_operation=BACKUP_OPERATION_NAME_RESTORE,
                )
                self.success_count += num_completed - len(new_anomalies)
                self.anomalies.extend(new_anomalies)
                self.anomalies.extend(self._subprocess_pipeline.anomalies)
            self._subprocess_pipeline.shutdown()

    def _get_unique_discovery_paths(self):
        # Create a list with unique discovery paths of the entire set of selected files.
        # Drop the drive letter and initial separator, while adding a add a trailing
        # separator for the 'startswith' check further below.
        all_disc_paths: set[str] = set(
            [
                f"{os.path.normcase(os.path.splitdrive(fi.discovery_path)[1][1:])}{os.path.sep}"
                for fi in self._selected_files
            ]
        )
        all_disc_paths = sorted(all_disc_paths)
        sel_disc_paths: list[str] = []
        cur_disc_path = None
        # Retain only the shortest path of all discovery paths.
        for d_path in all_disc_paths:
            if cur_disc_path is None or not d_path.startswith(cur_disc_path):
                cur_disc_path = d_path
                sel_disc_paths.append(d_path)
                continue
        return sel_disc_paths

    def _perform_restore_path_auto_mapping(self):
        # pylint: disable=unnecessary-lambda
        re_split_sep = re.compile(rf"[\\\/\{os.path.sep}]")
        sel_disc_paths = self._get_unique_discovery_paths()
        if len(sel_disc_paths) == 1:
            sel_disc_path = sel_disc_paths[0]
            num_to_remove = max(len(re_split_sep.split(sel_disc_path)) - 1, 0)
        else:
            sel_disc_paths_parts = list(map(re_split_sep.split, sel_disc_paths))
            min_part_count = min(
                map(
                    lambda pa: len(pa),
                    sel_disc_paths_parts,  # pylint: disable=unnecessary-lambda
                )  # pylint: disable=unnecessary-lambda
            )  # pylint: disable=unnecessary-lambda
            part_ele_idx = 0
            fin = False
            while not fin and part_ele_idx < min_part_count:
                v = sel_disc_paths_parts[0][part_ele_idx]
                for pa in sel_disc_paths_parts[1:]:
                    if v == pa[part_ele_idx]:
                        part_ele_idx += 1
                        continue
                    fin = True
                    break
            num_to_remove = part_ele_idx
        for fi in self._selected_files:
            parts = re_split_sep.split(fi.path_without_root)
            parts = parts[num_to_remove:]
            fi.restore_path_override = os.path.join(*parts)

    def restore_files(self):
        if self._auto_path_mapping:
            self._perform_restore_path_auto_mapping()
        self._restore_files()
        logging.info(f"All restore file operations have completed.")
        if len(self.anomalies) == 0:
            logging.info(f"***************")
            logging.info(f"*** SUCCESS ***")
            logging.info(f"***************")
            logging.info(f"No errors detected during restore.")
        else:
            log_anomalies_report(anomalies=self.anomalies)
        logging.info(f"{'Total files ':.<45} {len(self._selected_files)}")
        logging.info(f"{'Total errors ':.<45} {len(self.anomalies)}")
        logging.info(f"{'Total success ':.<45} {self.success_count}")


def restore_from_storage(
    storage_def: StorageDefinition,
    selections: list[BackupFileInformation],
    dest_root_location: str,
    allow_overwrite: bool,
    auto_path_mapping: bool,
) -> bool:
    with Restore(
        storage_def=storage_def,
        selections=selections,
        dest_root_location=dest_root_location,
        allow_overwrite=allow_overwrite,
        auto_path_mapping=auto_path_mapping,
    ) as restore:
        restore.restore_files()
    return not restore.is_failed


def handle_decrypt(args):
    storage_def_name: str = args.private_key_storage_specifier
    source_files_dir: str = args.source_files_dir
    dest_root_location: str = args.restore_dir
    allow_overwrite = args.overwrite

    storage_atbu_cfg: AtbuConfig
    storage_def_dict: dict = None
    if not is_existing_filesystem_storage_path(storage_location=storage_def_name):
        #
        # Storage definition is not a filesystem storage.
        # It must therefore be a storage-def-specifier or
        # simply/directly the storage_def_name itself.
        # Resolve to a storage_def_name...
        #
        parsed_name = parse_storage_def_specifier(storage_location=storage_def_name)
        if parsed_name:
            storage_def_name = parsed_name
        storage_atbu_cfg, _, _ = AtbuConfig.access_cloud_storage_config(
            storage_def_name=storage_def_name,
            must_exist=True,
            create_if_not_exist=False,
        )
        storage_def_dict = (
            storage_atbu_cfg.get_storage_def_with_resolved_secrets_deep_copy(
                storage_def_name=storage_def_name,
                keep_secrets_base64_encoded=False,
            )
        )
    else:
        #
        # Filesystem storage.
        #
        (
            storage_atbu_cfg,
            storage_def_name,
            storage_def_dict,
        ) = AtbuConfig.access_filesystem_storage_config(
            storage_location_path=storage_def_name,
            resolve_storage_def_secrets=True,
            create_if_not_exist=False,
            prompt_to_create=False,
        )

    storage_def = StorageDefinition.storage_def_from_dict(
        storage_def_name=storage_def_name,
        storage_def_dict=storage_def_dict,
    )

    source_files_dir = ensure_glob_pattern_for_dir(dir_wc=source_files_dir)

    path_list = glob.glob(pathname=source_files_dir, recursive=True)

    if len(path_list) == 0:
        raise InvalidCommandLineArgument(f"No files found for '{source_files_dir}'.")

    path_list = list(
        filter(
            lambda p: os.path.isfile(p)
            and (
                p.endswith(ATBU_FILE_BACKUP_EXTENSION_ENCRYPTED)
                or p.endswith(ATBU_FILE_BACKUP_EXTENSION)
            ),
            path_list,
        )
    )

    if len(path_list) == 0:
        raise InvalidCommandLineArgument(
            f"No {ATBU_FILE_BACKUP_EXTENSION_ENCRYPTED} or "
            f"{ATBU_FILE_BACKUP_EXTENSION} files found for "
            f"'{source_files_dir}'."
        )

    selections = list(
        map(
            lambda p: BackupFileInformation(path=p),
            path_list,
        )
    )

    for fi in selections:
        fi.is_decrypt_operation = True
        fi.populate_from_header = True

    is_this_successful = restore_from_storage(
        storage_def=storage_def,
        selections=selections,
        dest_root_location=dest_root_location,
        allow_overwrite=allow_overwrite,
        auto_path_mapping=False,
    )
    if is_this_successful:
        logging.info(f"Finished... no errors detected.")
        return 0
    else:
        logging.info(f"Finished... errors were detected.")
        return 1


def handle_restore(args):
    source_specifiers: str = args.source_storage_specifiers
    dest_root_location: str = args.restore_dir
    allow_overwrite = args.overwrite
    auto_path_mapping = args.auto_mapping

    sbs_list_list = user_specifiers_to_selections(specifiers=source_specifiers)
    for sbs_list in sbs_list_list:
        sbs_fi_count = 0
        for sbs in sbs_list:
            sbs_fi_count += len(sbs.selected_fi)
        logging.info(
            f"Will restore {sbs_fi_count} files from '{sbs_list[0].storage_def_name}'"
        )

    # No selections is failure.
    is_overall_success = len(sbs_list_list) > 0
    for sbs_list in sbs_list_list:

        verify_specific_backup_selection_list(sbs_list=sbs_list)
        if not isinstance(sbs_list[0].storage_def, StorageDefinition):
            raise InvalidStateError(f"The storage_def must be a StorageDefinition.")
        storage_def = sbs_list[0].storage_def
        selections = get_all_specific_backup_file_info(sbs_list=sbs_list)

        is_this_successful = restore_from_storage(
            storage_def=storage_def,
            selections=selections,
            dest_root_location=dest_root_location,
            allow_overwrite=allow_overwrite,
            auto_path_mapping=auto_path_mapping,
        )
        if not is_this_successful:
            is_overall_success = False
    if is_overall_success:
        logging.info(f"Finished... no errors detected.")
        return 0
    else:
        logging.info(f"Finished... errors were detected.")
        return 1



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Credential and storage-definition-related creation/management.
"""

import json
import os
import logging
import re
from send2trash import send2trash

from atbu.common.util_helpers import get_trash_bin_name, prompt_YN
from atbu.mp_pipeline.mp_global import switch_to_non_queued_logging

from .exception import *
from .constants import *
from .config import (
    AtbuConfig,
    get_automated_testing_mode,
    is_existing_filesystem_storage_path,
    is_storage_def_name_ok,
    parse_storage_def_specifier,
)
from .storage_def_credentials import StorageDefCredentialSet
from .credentials import (
    Credential,
    CredentialAesKey,
    CredentialByteArray,
    DescribedCredential,
    prompt_for_password,
    prompt_for_password_with_yubikey_opt,
)
from .storage_interface.base import (
    AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR,
    StorageInterfaceFactory,
)


def password_prompt(
    what: str = None,
    hidden: bool = True,
    allow_blank: bool = False,
) -> CredentialByteArray:

    if what is None:
        what = "password"

    def _is_password_valid(password: CredentialByteArray) -> bool:
        MAX_SIZE_IN_BYTES = 500  # Arbitrary, big enough.
        if len(password) > MAX_SIZE_IN_BYTES:
            print(f"The {what} you entered is too long.")
            print(f"The maximum length is {MAX_SIZE_IN_BYTES} UTF-8 encoded bytes.")
            return False
        if not allow_blank and len(password) == 0:
            print("Blank passwords are not allowed, try again.")
            return False
        return True

    if get_automated_testing_mode():
        # When automated tests are active, use Python's 'input' function which accepts standard
        # input from the test harness. This allows automated E2E testing of paths using this
        # function, where platform-specific console handling for hidden password input does not
        # utilize standard input.
        hidden = False

    return prompt_for_password(
        prompt=f"Enter {what}:",
        hidden=hidden,
        is_password_valid_func=_is_password_valid,
    )


def prompt_for_driver_storage_secrets(
    driver_kv_pairs: dict,
    secrets_visible: bool,
) -> dict:
    if CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY not in driver_kv_pairs:
        access_key = password_prompt(
            what="the cloud storage access key",
            hidden=not secrets_visible,
        )
        driver_kv_pairs[CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY] = access_key.decode(encoding="utf-8")

    if CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET not in driver_kv_pairs:
        secret_access_key = password_prompt(
            what="the cloud storage secret access key (aka \"secret\")",
            hidden=not secrets_visible,
        )
        driver_kv_pairs[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET] = secret_access_key


def backup_password_prompt_wizard() -> CredentialByteArray:
    print(
        """
You can require the backup to ask for a password before starting a backup/restore,
or you can allow a backup to proceed automatically without requiring your password.

When you choose the automatic approach which does not require a password, you are
allowing your backup 'private key' to be used automatically by this program. When
doing this, your backup private key is stored in a manner where, not only this
program, but other programs and people who have access to your computer or its
contents may be able to access and use your private key.

You can switch between requiring your password or using the automatic approach as
needed/desired. Regardless of your choice, you should be certain to back up your
security information (i.e., private key, related info) which you can do at any time.

"""
    )
    a = prompt_YN(
        prompt_msg="Choose whether to require password or not.",
        prompt_question=(
            "Require a (p)assword or allow (a)utomatic use of your backup's private key? "
        ),
        default_enter_ans="a",
        choices={"p": "[P/a]", "a": "[p/A]"},
    )

    if a == "a":
        return None
    else:
        print(
            """
You have chosen to require a password before a backup/restore can begin which requires you
to enter a password.
"""
        )
        password = prompt_for_password_with_yubikey_opt(
            prompt="Enter a password for this backup:",
            prompt_again="Enter a password for this backup again:",
        )
    return password


def setup_backup_encryption_wizard(
    storage_atbu_cfg: AtbuConfig, storage_def_name: str
) -> DescribedCredential:
    """Setup a password for the backup encryption credential if desired by user.

    Args:
        storage_atbu_cfg (AtbuConfig): The configuration to use.
        storage_def_name (str): The storage definition name.

    Raises:
        EncryptionAlreadyEnabledError: Encryption already enabled for the specified
            storage definition configuration.

    Returns:
        DescribedCredential: The DescribedCredential with a credential attribute
            that has its password in the clear if a password was specified.
    """
    a = prompt_YN(
        prompt_msg="The destination can be encrypted.",
        prompt_question="Would you like encryption enabled?",
    )
    credential: CredentialAesKey = None
    if a != "y":
        print(f"Files backed up to this destination will *not* be encrypted.")
        return None

    if storage_atbu_cfg.is_storage_def_configured_for_encryption(
        storage_def_name=storage_def_name
    ):
        raise EncryptionAlreadyEnabledError(
            f"Encryption is already enabled for storage_def={storage_def_name}."
        )
    cba_password = backup_password_prompt_wizard()
    print(f"Creating backup encryption key...", end="")
    credential = CredentialAesKey()
    credential.create_key()
    print(f"created.")
    if cba_password is not None:
        print(f"Setting password.")
        credential.set(password=cba_password)
    else:
        print(
            f"No password set because you want to keep the backup encryption key unencrypted."
        )

    return DescribedCredential(
        credential=credential,
        config_name=storage_def_name,
        credential_name=CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
        credential_kind=CONFIG_PASSWORD_KIND_ACTUAL,
    )


def is_OAuth2_secret_json_file(path: str) -> bool:
    if path is None or not os.path.isfile(path):
        return False
    # There is a candidate file, see if it ends with .json and contains expected fields...
    logging.info(
        f"Secret seems to reference a file either directly or indirectly: {path}"
    )
    if not path.lower().endswith(".json"):
        logging.warning(f"No .json extension, will not use file as secret: {path}")
        return False
    try:
        with open(path, "r", encoding="utf-8") as cred_file:
            c: dict = json.load(cred_file)
        if c.get("type") is None and c.get("private_key") is None:
            logging.warning(
                f"No 'type' or 'private_key' fields found, will not use file as secret: {path}"
            )
            return False
        # It appears to be a file, such as service account file (i.e., with OAuth2 creds)...
        logging.info(f"Secret will be considered a reference to a file: {path}")
        return True
    except Exception:
        logging.warning(f"Cannot open file, will not use file as secret: {path}")
        return False


def handle_container_auto_creation(
    storage_def_name: str, try_create_if_not_autofind: bool
) -> bool:
    atbu_config, _, storage_def_dict = AtbuConfig.access_cloud_storage_config(
        storage_def_name=storage_def_name,
        must_exist=True,
        create_if_not_exist=False,
    )
    container_name = storage_def_dict[CONFIG_VALUE_NAME_CONTAINER]
    auto_find_name = container_name[-1] == AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR
    if not auto_find_name and not try_create_if_not_autofind:
        # Auto find implies a need to create a container, whereas just
        # the name requires --create-container. If neither auto find
        # nor --create-container, return.
        return False
    if auto_find_name:
        logging.info(
            f"Container name had the {AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR} "
            f"auto-find/create indicator. "
            f"Searching for unique container name using base name {container_name}..."
        )
    factory = StorageInterfaceFactory.create_factory_from_storage_def_name(
        storage_def_name=storage_def_name
    )
    container_name = factory.storage_def_dict[CONFIG_VALUE_NAME_CONTAINER]
    interface = factory.create_storage_interface()
    try:
        ci = interface.create_container(container_name=container_name)
        logging.info(f"Found/created container name '{ci.name}'.")
        logging.info(f"Updating configuration with that new name.")
    except ContainerAlreadyExistsError:
        logging.warning(
            f"WARNING: The container name chosen already exists: {container_name}"
        )
        logging.warning(
            f"If you do not intend to use this container name, try againg with a different name."
        )
        logging.info(f"Using container name '{container_name}'.")
        logging.info(f"Updating configuration with that existing name.")
    except InvalidContainerNameError:
        logging.error(f"The container name is invalid: {container_name}")
        raise
    except ContainerAutoCreationFailed as ex:
        logging.error(exc_to_string(ex))
        logging.error(
            f"Auto find container with creation failed for name '{container_name}' "
            f"Check the errors and the name."
        )
        raise
    storage_def_dict[CONFIG_VALUE_NAME_CONTAINER] = ci.name
    atbu_config.save_config_file()
    return True


def handle_create_storage_definition(
    storage_def_name: str,
    interface: str,
    provider: str,
    container: str,
    driver_params: str,
    create_container: bool,
    include_iv: bool,
    secrets_visible: bool,
):
    # Disable MP queued logging to ensure any console
    # logging is in sync with print statements.
    switch_to_non_queued_logging()

    if not storage_def_name:
        raise ValueError(
            f"Invalid name storage_def_name '{storage_def_name}' specified."
        )
    storage_def_name = storage_def_name.lower()
    if not interface:
        raise ValueError(f"Invalid name interface '{interface}' specified.")
    if not provider:
        raise ValueError(f"Invalid name provider '{provider}' specified.")
    if not container:
        raise ValueError(f"Invalid name container '{container}' specified.")

    atbu_cfg: AtbuConfig
    atbu_cfg, _, _ = AtbuConfig.access_cloud_storage_config(
        storage_def_name=storage_def_name,
        must_exist=False,
        create_if_not_exist=False,
    )

    if atbu_cfg is not None:
        raise StorageDefinitionAlreadyExists(
            f"The storage definition '{storage_def_name}' already exists. "
            f"Delete it first, or choose a different name."
        )

    if not is_storage_def_name_ok(storage_def_name=storage_def_name):
        raise InvalidStorageDefinitionName(
            f"The name '{storage_def_name}' is invalid. Allowed characters: "
            f"lowercase alphanumeric, underscore, hyphen/dash."
        )

    atbu_cfg = AtbuConfig.create_cloud_storage_def_config(
        storage_def_name=storage_def_name,
    )

    if driver_params is None:
        driver_kv_pairs = {}
    else:
        secret = None
        m = re.match(r".*[,\s]*secret=([^,]+)", driver_params)
        if m:
            # A base64 secret may contain '=' so remove and parse separately.
            secret = CredentialByteArray(m.groups()[0].encode("utf-8"))
            driver_params = re.sub(r"[,\s]*secret=([^,]+)", "", driver_params).strip(', ')
        driver_kv_pairs = dict(
            kv_pair.split("=") for kv_pair in driver_params.split(",")
        )
        if secret is not None:
            driver_kv_pairs[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET] = secret

    if (
        CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY in driver_kv_pairs
        or CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET in driver_kv_pairs
    ):
        print(f"WARNING: You have specified secrets on the command line which means")
        print(f"those secrets may be stored in your command line history and more easily")
        print(f"accessible to other parties accessing your login. You should consider")
        print(f"inputing the secrets directly by not specifying them on the command line, or")
        print(f"securely clearing, or disabling your command line history file.")

    prompt_for_driver_storage_secrets(
        driver_kv_pairs=driver_kv_pairs,
        secrets_visible=secrets_visible
    )

    _, storage_def_dict = atbu_cfg.create_storage_def(
        interface_type=interface,
        provider_id=provider,
        container=container,
        other_driver_kv_pairs=driver_kv_pairs,
        unique_storage_def_name=storage_def_name,
        include_iv=include_iv,
    )

    #
    # Determine storage credential type (i.e., password, file, or env var file).
    #

    cred_storage_kind = CONFIG_PASSWORD_KIND_ACTUAL
    secret = driver_kv_pairs[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET]
    secret_str = secret.decode("utf-8")
    if os.path.isfile(secret_str) and is_OAuth2_secret_json_file(path=secret_str):
        # Yes, secret is directly an existing file name.
        cred_storage_kind = CONFIG_PASSWORD_KIND_FILENAME
    else:
        # Not a direct file name.
        # Check if env var.
        ev = os.getenv(secret_str)
        if isinstance(ev, str) and is_OAuth2_secret_json_file(path=ev):
            # Yes, env var is file.
            cred_storage_kind = CONFIG_PASSWORD_KIND_ENVVAR

    #
    # Create/save the credentials.
    #

    try:

        #
        # Create the backup encryption credential.
        # If no encryption, desc_cred_encryption is None.
        #

        desc_cred_encryption = setup_backup_encryption_wizard(
            storage_atbu_cfg=atbu_cfg,
            storage_def_name=storage_def_name,
        )

        #
        # Create the storage credential.
        # If there is an encryption credential, and it is password-protected,
        # use the same password to protect the storage credential.
        #

        cred_storage = Credential(the_key=secret)
        if desc_cred_encryption is not None:
            cred_encryption = desc_cred_encryption.credential
            if cred_encryption.is_password_protected:
                cred_storage.set(password=CredentialByteArray(cred_encryption.password))

        desc_cred_storage = DescribedCredential(
            credential=cred_storage,
            config_name=storage_def_name,
            credential_name=CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD,
            credential_kind=cred_storage_kind,
        )

        #
        # Create a credential set to hold/lock/save all credentials.
        #

        cred_set = StorageDefCredentialSet(
            storage_def_name=storage_def_name,
            storage_def_dict=storage_def_dict,
        )
        if desc_cred_encryption is not None:
            cred_set.append(
                desc_cred=desc_cred_encryption,
                affected_config_path_parts=CRED_SECRET_KIND_ENCRYPTION.split("-"),
            )
        cred_set.append(
            desc_cred=desc_cred_storage,
            affected_config_path_parts=CRED_SECRET_KIND_STORAGE.split("-"),
        )

        #
        # Lock and save all credentials.
        #

        print(f"Storing...")
        cred_set.protect()
        cred_set.save()
        print(f"Credentials stored.")
        print(f"Saving {atbu_cfg.path}")
        atbu_cfg.save_config_file()
        logging.info(f"Storage definition {storage_def_name} saved.")
        # Test log line: The following is used by tests:
        logging.debug(f"Storage definition {storage_def_name} saved to {atbu_cfg.path}")

        if desc_cred_encryption is None:
            logging.warning(
                f"WARNING: The storage definition '{storage_def_name}' will *not* be encrypted."
            )
        else:
            logging.info(
                f"The storage definition '{storage_def_name}' will be encrypted."
            )

        handle_container_auto_creation(
            storage_def_name=storage_def_name,
            try_create_if_not_autofind=create_container,
        )

        logging.info(f"Storage definition {storage_def_name} successfully created.")
    except Exception as ex:
        message = str(ex)
        if isinstance(ex, AtbuException):
            message = ex.message
        if atbu_cfg.is_storage_def_exists(storage_def_name=storage_def_name):
            logging.warning(
                f"Due to error, deleting storage definition '{storage_def_name}'. "
                f"{message} {exc_to_string(ex)}"
            )
            atbu_cfg.delete_storage_def(storage_def_name=storage_def_name)
            atbu_cfg.save_config_file()
        raise


def handle_delete_storage_definition(
    storage_def_name: str,
    skip_confirmation_prompt: bool,
    delete_backup_info: bool,
):
    """Delete a storage definition.

    Args:
        storage_def_name: The storage definition name.
        skip_confirmation_prompt: If True, then no
            prompting will occur, definition will be
            deleted.
        delete_backup_info: If True, the backup information
            will also be deleted, if False the backup
            information will not be deleted. If None,
            and skip_confirmation_prompt==False, the
            user will be prompted about backup information,
            otherwise in the case of not prompting, the
            backup information will not be deleted. The
            goal is to be cautious about deleting backup
            informoation despite it being recoverable from
            a good backup. Essentially, the user must be
            explicitly one way or another about including
            backup information in the deletion process.
    """
    switch_to_non_queued_logging()
    bin_name = get_trash_bin_name()

    atbu_cfg: AtbuConfig
    atbu_cfg, _, _ = AtbuConfig.access_cloud_storage_config(
        storage_def_name=storage_def_name,
        must_exist=False,
        create_if_not_exist=False,
        storage_def_dict_not_exist_ok=True,
    )
    if atbu_cfg is None:
        print(f"The storage definition '{storage_def_name}' does not exist.")
        return

    existing_backup_info = atbu_cfg.get_primary_backup_info_file_paths(
        storage_def_name=storage_def_name
    )

    if not skip_confirmation_prompt:
        print(
            f"""*** WARNING *** WARNING *** WARNING *** WARNING *** WARNING *** WARNING
The storage definition '{storage_def_name}' exists.
If this is an encrypted backup where the private key is not backed up,
you *will* *lose* *access* to all data in this backup if you delete this
configuration.
"""
        )
        if (
            delete_backup_info is not None  # If user was explicit
            and delete_backup_info  # and user said Yes to deleting BI
            and len(existing_backup_info) > 0  # and there is BI to delete
        ):
            print()
            print(
                f"""If you delete this storage definition, the following backup information
will also be deleted:
"""
            )
            print()
            for bip in existing_backup_info:
                print(f"    {bip}")
            print()
        a = prompt_YN(
            prompt_msg=f"You are about to delete a backup storage definition.",
            prompt_question=f"Are you certain you want to delete '{storage_def_name}' ",
            default_enter_ans="n",
        )
        if a != "y":
            print("The storage definition will not be deleted.")
            return
        if (
            delete_backup_info is None  # user was not explicit
            and len(existing_backup_info) > 0  # there is BI to delete
        ):
            print()
            print(f"This storage definition also has backup information as follows:")
            print()
            for bip in existing_backup_info:
                print(f"    {bip}")
            print()
            print(
                f"""If you are planning to immediately re-import this same configuration,
or you otherwise need this backup information, you should not delete it.

While this backup information is typically recoverable from a non-corrupt
backup, it is generally a good idea to err on the side of caution and back
it up if you are uncertain.

If you choose 'y' below to delete the above files, they will no longer be
available. This app will attempt to send the files to the {bin_name},
so you may be able to recover them from there if needed. Choosing 'y' will
only delete backup information files in the primary backup information
directory. Information files in the secondary backup information locations,
if any, will not be deleted (you will have to do so manually, if desired).

If you choose *not* to delete these files along with the storage definition,
and you later attempt import/recovery of a storage definition of the same name,
the lingering presence of these older files may interfere or be out of sync with
that newly imported configuration.
                """
            )
            a = prompt_YN(
                prompt_msg=(
                    f"Choose whether to also delete the above backup information files (a=abort)."
                ),
                prompt_question=(
                    f"Delete backup information files along with '{storage_def_name}' "
                ),
                default_enter_ans="a",
                choices={"y": "[Y/n/a]", "n": "[y/N/a]", "a": "[y/n/A]"},
            )
            if a == "a":
                print("The storage definition will not be deleted.")
                return
            delete_backup_info = a == "y"
    logging.info(f"Deleting storage definition '{storage_def_name}'...")
    atbu_cfg.delete_storage_def(storage_def_name=storage_def_name)
    atbu_cfg.delete_config_file()
    if delete_backup_info:
        logging.info("Deleting backup information...")
        for bip in existing_backup_info:
            logging.info(f"    Sending file to {bin_name}: {bip}")
            send2trash(bip)
    elif len(existing_backup_info) > 0:
        logging.info(f"Skipping deletion of the following backup information...")
        for bip in existing_backup_info:
            logging.info(f"    {bip}")
    else:
        logging.info(f"There is no backup information for '{storage_def_name}'.")
    logging.info(f"The storage definition '{storage_def_name}' was deleted.")


def handle_creds(args):
    logging.debug(f"handle_creds")
    storage_def_name = args.storage_def
    orig_storage_def_name = storage_def_name  # pylint: disable=unused-variable
    if args.subcmd == CREDS_SUBCMD_CREATE_STORAGE_DEF:
        handle_create_storage_definition(
            storage_def_name=storage_def_name,
            interface=args.interface,
            provider=args.provider,
            container=args.container,
            driver_params=args.driver_params,
            create_container=args.create_container,
            include_iv=args.include_iv,
            secrets_visible=args.secrets_visible,
        )
    elif args.subcmd == "delete-storage-def":
        handle_delete_storage_definition(
            storage_def_name=storage_def_name,
            skip_confirmation_prompt=args.force,
            delete_backup_info=args.delete_backup_info,
        )
    elif args.subcmd == "export":
        storage_atbu_cfg: AtbuConfig
        if not is_existing_filesystem_storage_path(storage_location=storage_def_name):
            #
            # Storage definition is not a filesystem storage.
            # It must therefore be a storage-def-specifier or
            # simply/directly the storage_def_name itself.
            # Resolve to a storage_def_name...
            #
            parsed_name = parse_storage_def_specifier(storage_location=storage_def_name)
            if parsed_name is not None:
                storage_def_name = parsed_name
            storage_atbu_cfg, _, _ = AtbuConfig.access_cloud_storage_config(
                storage_def_name=storage_def_name,
                must_exist=False,
                create_if_not_exist=False,
            )
        else:
            #
            # Filesystem storage.
            #
            storage_atbu_cfg, storage_def_name, _ = AtbuConfig.access_filesystem_storage_config(
                storage_location_path=storage_def_name,
                resolve_storage_def_secrets=False,
                create_if_not_exist=True,
                prompt_to_create=True,
            )
        if storage_atbu_cfg is None:
            raise StorageDefinitionNotFoundError(
                f"The storage definition '{storage_def_name}' was not found."
            )
        backup_file_path = args.filename
        if not backup_file_path:
            raise InvalidCommandLineArgument(
                f"You must specify a filename to save the storage definition to."
            )
        if os.path.isdir(backup_file_path):
            raise InvalidCommandLineArgument(
                f"The specified file is a directory: {backup_file_path}"
            )
        if os.path.isfile(backup_file_path) and not args.overwrite:
            raise InvalidCommandLineArgument(
                f"The file already exists: {backup_file_path}"
            )
        storage_atbu_cfg.save_storage_def_with_cleartext_secrets(
            storage_def_name=storage_def_name, backup_file_path=backup_file_path
        )
    elif args.subcmd == "import":
        storage_atbu_cfg: AtbuConfig
        if not is_existing_filesystem_storage_path(storage_location=storage_def_name):
            #
            # Storage definition is not a filesystem storage.
            # It must therefore be a storage-def-specifier or
            # simply/directly the storage_def_name itself.
            # Resolve to a storage_def_name...
            #
            parsed_name = parse_storage_def_specifier(storage_location=storage_def_name)
            if parsed_name is not None:
                storage_def_name = parsed_name
            storage_atbu_cfg, _, _ = AtbuConfig.access_cloud_storage_config(
                storage_def_name=storage_def_name,
                must_exist=False,
                create_if_not_exist=False,
                storage_def_dict_not_exist_ok=True,
            )
        else:
            #
            # Filesystem storage.
            #
            storage_atbu_cfg, storage_def_name, _ = AtbuConfig.access_filesystem_storage_config(
                storage_location_path=storage_def_name,
                resolve_storage_def_secrets=False,
                create_if_not_exist=True,
                prompt_to_create=True,
            )
        switch_to_non_queued_logging()
        # TODO: Handle args.create_config
        backup_file_path = args.filename
        if not backup_file_path:
            raise InvalidCommandLineArgument(
                f"You must specify the filename of the storage definition and "
                f"credentials to restore."
            )
        if os.path.isdir(backup_file_path):
            raise InvalidCommandLineArgument(
                f"The specified file is a directory: {backup_file_path}"
            )
        if not os.path.isfile(backup_file_path):
            raise InvalidCommandLineArgument(
                f"The file does not exist: {backup_file_path}"
            )

        if storage_atbu_cfg is None:
            storage_atbu_cfg, _, _ = AtbuConfig.access_cloud_storage_config(
                storage_def_name=storage_def_name,
                must_exist=False,
                create_if_not_exist=True,
            )
        storage_atbu_cfg.restore_storage_def(
            storage_def_new_name=storage_def_name,
            backup_file_path=backup_file_path,
            prompt_if_exists=args.prompt,
        )
    elif args.subcmd in [
        CRED_OPERATION_SET_PASSWORD,
        CRED_OPERATION_SET_PASSWORD_ALIAS,
    ]:
        storage_atbu_cfg: AtbuConfig
        if not is_existing_filesystem_storage_path(storage_location=storage_def_name):
            #
            # Storage definition is not a filesystem storage.
            # It must therefore be a storage-def-specifier or
            # simply/directly the storage_def_name itself.
            # Resolve to a storage_def_name...
            #
            parsed_name = parse_storage_def_specifier(storage_location=storage_def_name)
            if parsed_name is not None:
                storage_def_name = parsed_name
            storage_atbu_cfg, _, _ = AtbuConfig.access_cloud_storage_config(
                storage_def_name=storage_def_name,
                must_exist=False,
                create_if_not_exist=False,
            )
        else:
            #
            # Filesystem storage.
            #
            storage_atbu_cfg, storage_def_name, _ = AtbuConfig.access_filesystem_storage_config(
                storage_location_path=storage_def_name,
                resolve_storage_def_secrets=False,
                create_if_not_exist=True,
                prompt_to_create=True,
            )

        switch_to_non_queued_logging()

        if storage_atbu_cfg is None:
            raise StorageDefinitionNotFoundError(
                f"The storage definition '{storage_def_name}' was not found."
            )

        storage_def_dict = storage_atbu_cfg.get_storage_def_dict(
            storage_def_name=storage_def_name,
            must_exist=True,
        )

        cred_set = StorageDefCredentialSet(
            storage_def_name=storage_def_name,
            storage_def_dict=storage_def_dict,
        )

        cred_set.populate()
        cred_set.unprotect()

        password_type = args.password_type

        if password_type == CRED_OPERATION_SET_PASSWORD_TYPE_BACKUP:
            if args.password is not None:
                raise CredentialRequestInvalidError(
                    f"A command line password cannot be specified for the "
                    f"{CRED_OPERATION_SET_PASSWORD_TYPE_BACKUP} password."
                )
            cba_password = backup_password_prompt_wizard()
            cred_set.set_password(cba_password)
        elif password_type == CRED_OPERATION_SET_PASSWORD_TYPE_STORAGE:
            desc_cred_storage = cred_set.get_storage_desc_cred()
            if args.password is None:
                cba_password = password_prompt(what="storage secret", hidden=False)
            else:
                cba_password = CredentialByteArray.create_from_string(args.password)
            desc_cred_storage.credential.set(the_key=cba_password)
            desc_cred_storage.credential_kind = CONFIG_PASSWORD_KIND_ACTUAL
        elif password_type == CRED_OPERATION_SET_PASSWORD_TYPE_FILENAME:
            desc_cred_storage = cred_set.get_storage_desc_cred()
            if args.password is None:
                cba_password = password_prompt(
                    what="OAuth2 .json file path",
                    hidden=False,
                )
            else:
                cba_password = CredentialByteArray.create_from_string(args.password)
            desc_cred_storage.credential.set(the_key=cba_password)
            desc_cred_storage.credential_kind = CONFIG_PASSWORD_KIND_FILENAME
        elif password_type == CRED_OPERATION_SET_PASSWORD_TYPE_ENVVAR:
            desc_cred_storage = cred_set.get_storage_desc_cred()
            if args.password is None:
                cba_password = password_prompt(
                    what="OAuth2 .json file path environment variable",
                    hidden=False,
                )
            else:
                cba_password = CredentialByteArray.create_from_string(args.password)
            desc_cred_storage.credential.set(the_key=cba_password)
            desc_cred_storage.credential_kind = CONFIG_PASSWORD_KIND_ENVVAR
        else:
            raise InvalidStateError(f"Unexpected password type '{password_type}'")

        print(f"Storing credentials...")
        cred_set.protect()
        cred_set.save()
        print(f"Credentials stored.")
    else:
        raise InvalidCommandLineArgument("Unknown command.")
